from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITranslationManager
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.dexterity.utils import createContentInContainer
from plone.restapi.interfaces import IExpandableElement
from plone.restapi.serializer.expansion import expandable_elements
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from zope.component import getGlobalSiteManager
from zope.component import provideAdapter
from zope.interface import alsoProvides
from zope.interface import Interface
from zope.publisher.browser import TestRequest
from zope.publisher.interfaces.browser import IBrowserRequest

import transaction
import unittest


class ExpandableElementFoo:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, expand=False):
        if expand:
            return {"foo": "expanded"}
        else:
            return {"foo": "collapsed"}


class ExpandableElementBar:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, expand=False):
        if expand:
            return {"bar": "expanded"}
        else:
            return {"bar": "collapsed"}


class TestExpansion(unittest.TestCase):
    def setUp(self):
        provideAdapter(
            ExpandableElementFoo,
            adapts=(Interface, IBrowserRequest),
            provides=IExpandableElement,
            name="foo",
        )
        provideAdapter(
            ExpandableElementBar,
            adapts=(Interface, IBrowserRequest),
            provides=IExpandableElement,
            name="bar",
        )

    def test_expansion_returns_collapsed_elements(self):
        request = TestRequest()
        self.assertEqual(
            {"@components": {"bar": "collapsed", "foo": "collapsed"}},
            expandable_elements(None, request),
        )

    def test_expansion_returns_expanded_element(self):
        request = TestRequest(form={"expand": "foo"})
        self.assertEqual(
            {"@components": {"bar": "collapsed", "foo": "expanded"}},
            expandable_elements(None, request),
        )

    def test_expansion_returns_multiple_deprecated_commas_expanded_elements(self):
        request = TestRequest(form={"expand": "foo,bar"})
        self.assertEqual(
            {"@components": {"bar": "expanded", "foo": "expanded"}},
            expandable_elements(None, request),
        )

    def test_expansion_returns_multiple_expanded_elements(self):
        request = TestRequest(form={"expand": ["foo", "bar"]})
        self.assertEqual(
            {"@components": {"bar": "expanded", "foo": "expanded"}},
            expandable_elements(None, request),
        )

    def tearDown(self):
        gsm = getGlobalSiteManager()
        gsm.unregisterAdapter(
            ExpandableElementFoo,
            (Interface, IBrowserRequest),
            IExpandableElement,
            "foo",
        )
        gsm.unregisterAdapter(
            ExpandableElementBar,
            (Interface, IBrowserRequest),
            IExpandableElement,
            "bar",
        )


class TestExpansionFunctional(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.folder = createContentInContainer(
            self.portal, "Folder", id="folder", title="Some Folder"
        )
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_expanded_marked_as_list(self):
        response = self.api_session.get(
            "/folder", params={"expand:list": ["actions", "navigation"]}
        )

        self.assertEqual(response.status_code, 200)
        self.assertTrue("object" in response.json()["@components"]["actions"])
        self.assertTrue("object_buttons" in response.json()["@components"]["actions"])
        self.assertTrue("portal_tabs" in response.json()["@components"]["actions"])
        self.assertTrue("site_actions" in response.json()["@components"]["actions"])
        self.assertTrue("user" in response.json()["@components"]["actions"])
        self.assertEqual(
            [
                {
                    "title": "Home",
                    "@id": self.portal_url + "",
                    "description": "",
                    "review_state": None,
                    "items": [],
                },
                {
                    "title": "Some Folder",
                    "@id": self.portal_url + "/folder",
                    "description": "",
                    "review_state": "private",
                    "items": [],
                },
            ],
            response.json()["@components"]["navigation"]["items"],
        )

    def test_expanded_as_list(self):
        response = self.api_session.get(
            "/folder", params={"expand": ["actions", "navigation"]}
        )

        self.assertEqual(response.status_code, 200)
        self.assertTrue("object" in response.json()["@components"]["actions"])
        self.assertTrue("object_buttons" in response.json()["@components"]["actions"])
        self.assertTrue("portal_tabs" in response.json()["@components"]["actions"])
        self.assertTrue("site_actions" in response.json()["@components"]["actions"])
        self.assertTrue("user" in response.json()["@components"]["actions"])
        self.assertEqual(
            [
                {
                    "title": "Home",
                    "@id": self.portal_url + "",
                    "description": "",
                    "review_state": None,
                    "items": [],
                },
                {
                    "title": "Some Folder",
                    "@id": self.portal_url + "/folder",
                    "description": "",
                    "review_state": "private",
                    "items": [],
                },
            ],
            response.json()["@components"]["navigation"]["items"],
        )

    def test_actions_is_expandable(self):
        response = self.api_session.get("/folder")

        self.assertEqual(response.status_code, 200)
        self.assertIn("actions", list(response.json().get("@components")))

    def test_actions_expanded(self):
        response = self.api_session.get("/folder", params={"expand": "actions"})

        self.assertEqual(response.status_code, 200)
        self.assertTrue("object" in response.json()["@components"]["actions"])
        self.assertTrue("object_buttons" in response.json()["@components"]["actions"])
        self.assertTrue("portal_tabs" in response.json()["@components"]["actions"])
        self.assertTrue("site_actions" in response.json()["@components"]["actions"])
        self.assertTrue("user" in response.json()["@components"]["actions"])

    def test_navigation_is_expandable(self):
        response = self.api_session.get("/folder")

        self.assertEqual(response.status_code, 200)
        self.assertIn("navigation", list(response.json().get("@components")))

    def test_navigation_expanded(self):
        response = self.api_session.get("/folder", params={"expand": "navigation"})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            [
                {
                    "title": "Home",
                    "@id": self.portal_url + "",
                    "description": "",
                    "review_state": None,
                    "items": [],
                },
                {
                    "title": "Some Folder",
                    "@id": self.portal_url + "/folder",
                    "description": "",
                    "review_state": "private",
                    "items": [],
                },
            ],
            response.json()["@components"]["navigation"]["items"],
        )

    def test_navigation_expanded_with_depth(self):
        createContentInContainer(
            self.portal, "Folder", id="folder2", title="Some Folder 2"
        )
        subfolder1 = createContentInContainer(
            self.folder, "Folder", id="subfolder1", title="SubFolder 1"
        )
        createContentInContainer(
            self.folder, "Folder", id="subfolder2", title="SubFolder 2"
        )
        thirdlevelfolder = createContentInContainer(
            subfolder1, "Folder", id="thirdlevelfolder", title="Third Level Folder"
        )
        createContentInContainer(
            thirdlevelfolder,
            "Folder",
            id="fourthlevelfolder",
            title="Fourth Level Folder",
        )
        createContentInContainer(self.folder, "Document", id="doc1", title="A document")
        transaction.commit()

        response = self.api_session.get(
            "/folder", params={"expand": "navigation", "expand.navigation.depth": 3}
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()["@components"]["navigation"]["items"]), 3)
        self.assertEqual(
            len(response.json()["@components"]["navigation"]["items"][1]["items"]),
            3,  # noqa
        )
        self.assertEqual(
            len(
                response.json()["@components"]["navigation"]["items"][1]["items"][0][
                    "items"
                ]
            ),
            1,  # noqa
        )

    def test_breadcrumbs_is_expandable(self):
        response = self.api_session.get("/folder")

        self.assertEqual(response.status_code, 200)
        self.assertIn("breadcrumbs", list(response.json().get("@components")))

    def test_breadcrumbs_expanded(self):
        response = self.api_session.get("/folder", params={"expand": "breadcrumbs"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            [
                {
                    "title": "Some Folder",
                    "@id": self.portal_url + "/folder",
                }
            ],
            response.json()["@components"]["breadcrumbs"]["items"],
        )

    def test_workflow_is_expandable(self):
        response = self.api_session.get("/folder")

        self.assertEqual(response.status_code, 200)
        self.assertIn("workflow", list(response.json().get("@components")))

    def test_workflow_expanded(self):
        response = self.api_session.get("/folder", params={"expand": "workflow"})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            self.portal_url + "/folder/@workflow",
            response.json().get("@components").get("workflow").get("@id"),
        )
        self.assertEqual(
            "private",
            response.json()["@components"]["workflow"]["history"][0][
                "review_state"
            ],  # noqa
        )
        self.assertEqual(
            "Private",
            response.json()["@components"]["workflow"]["history"][0]["title"],
        )
        self.assertEqual(
            [
                {
                    "@id": self.portal_url + "/folder/@workflow/publish",  # noqa
                    "title": "Publish",
                },
                {
                    "@id": self.portal_url + "/folder/@workflow/submit",  # noqa
                    "title": "Submit for publication",
                },
            ],
            response.json()["@components"]["workflow"]["transitions"],
        )

    def test_types_is_expandable(self):
        response = self.api_session.get("/folder")

        self.assertEqual(response.status_code, 200)
        self.assertIn("types", list(response.json().get("@components")))

    def test_types_expanded(self):
        response = self.api_session.get("/folder", params={"expand": "types"})

        self.assertEqual(response.status_code, 200)

        # XXX: Note: The @types endpoint currently doesn't conform to JSON-LD
        # because it's directly returning a list, and does not have an @id
        # property.

        base_url = self.portal.absolute_url()

        self.assertEqual(
            [
                {
                    "@id": "/".join((base_url, "@types/Collection")),
                    "addable": True,
                    "immediately_addable": True,
                    "title": "Collection",
                },
                {
                    "@id": "/".join((base_url, "@types/DXTestDocument")),
                    "addable": True,
                    "immediately_addable": True,
                    "title": "DX Test Document",
                },
                {
                    "@id": "/".join((base_url, "@types/Event")),
                    "addable": True,
                    "immediately_addable": True,
                    "title": "Event",
                },
                {
                    "@id": "/".join((base_url, "@types/File")),
                    "addable": True,
                    "immediately_addable": True,
                    "title": "File",
                },
                {
                    "@id": "/".join((base_url, "@types/Folder")),
                    "addable": True,
                    "immediately_addable": True,
                    "title": "Folder",
                },
                {
                    "@id": "/".join((base_url, "@types/Image")),
                    "addable": True,
                    "immediately_addable": True,
                    "title": "Image",
                },
                {
                    "@id": "/".join((base_url, "@types/Link")),
                    "addable": True,
                    "immediately_addable": True,
                    "title": "Link",
                },
                {
                    "@id": "/".join((base_url, "@types/News Item")),
                    "addable": True,
                    "immediately_addable": True,
                    "title": "News Item",
                },
                {
                    "@id": "/".join((base_url, "@types/Document")),
                    "addable": True,
                    "immediately_addable": True,
                    "title": "Page",
                },
            ],
            response.json().get("@components").get("types"),
        )


class TestTranslationExpansionFunctional(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        login(self.portal, SITE_OWNER_NAME)
        self.en_content = createContentInContainer(
            self.portal["en"], "Document", title="Test document"
        )
        self.en_folder = createContentInContainer(
            self.portal["en"], "Folder", title="Test folder"
        )
        self.es_content = createContentInContainer(
            self.portal["es"], "Document", title="Test document"
        )
        ITranslationManager(self.en_content).register_translation("es", self.es_content)

        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_translations_is_expandable(self):
        response = self.api_session.get("/en/test-document")

        self.assertEqual(response.status_code, 200)
        self.assertIn("translations", list(response.json().get("@components")))

    def test_translations_expanded(self):
        response = self.api_session.get(
            "/en/test-document", params={"expand": "translations"}
        )

        self.assertEqual(response.status_code, 200)
        translation_dict = {"@id": self.es_content.absolute_url(), "language": "es"}
        self.assertIn(
            translation_dict, response.json()["@components"]["translations"]["items"]
        )

    def test_expansions_no_fullobjects_do_not_modify_id(self):
        response = self.api_session.get(
            "/en/test-folder", params={"expand": "translations"}
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["@id"], self.en_folder.absolute_url())



from plone import api
from plone.app.discussion.interfaces import IDiscussionSettings
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.registry.interfaces import IRegistry
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from zope.component import getUtility

import transaction
import unittest


class TestCommentsEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.portal_url = self.portal.absolute_url()

        # Allow discussion
        registry = getUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True
        settings.edit_comment_enabled = True
        settings.delete_own_comment_enabled = True

        # doc with comments
        self.doc = api.content.create(
            container=self.portal,
            type="Document",
            id="doc_with_comments",
            title="Document with comments",
            allow_discussion=True,
        )
        api.content.transition(self.doc, "publish")

        api.user.create(username="jos", password="josjos", email="jos@plone.org")

        # Admin session
        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        # User session
        self.user_session = RelativeSession(self.portal_url, test=self)
        self.user_session.headers.update({"Accept": "application/json"})
        self.user_session.auth = ("jos", "jos")

        transaction.commit()

    def tearDown(self):
        self.api_session.close()
        self.user_session.close()

    def test_list_datastructure(self):
        url = f"{self.doc.absolute_url()}/@comments"
        response = self.api_session.get(url)

        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual({"items_total", "items", "@id"}, set(data))

    def test_list_batching(self):
        url = f"{self.doc.absolute_url()}/@comments"

        self.api_session.post(url, json={"text": "comment 1"})
        self.api_session.post(url, json={"text": "comment 2"})

        response = self.api_session.get(url, params={"b_size": 1})
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertIn("batching", data)

    def test_add_comment_to_root(self):
        url = f"{self.doc.absolute_url()}/@comments"

        response = self.api_session.get(url)
        self.assertEqual(0, response.json()["items_total"])

        response = self.api_session.post(url, json={"text": "comment 1"})
        self.assertEqual(204, response.status_code)
        self.assertIn("location", response.headers)

        response = self.api_session.get(url)
        data = response.json()
        self.assertEqual(1, data["items_total"])
        self.assertIsNone(data["items"][0]["in_reply_to"])
        self.assertIsNone(data["items"][0]["@parent"])

    def test_add_comment_to_comment(self):
        url = f"{self.doc.absolute_url()}/@comments"

        response = self.api_session.post(url, json={"text": "comment 1"})
        self.assertEqual(204, response.status_code)

        response = self.api_session.get(url)
        data = response.json()
        parent_id = data["items"][0]["comment_id"]

        SUBTEXT = "sub comment"

        payload = {"text": SUBTEXT, "in_reply_to": parent_id}
        response = self.api_session.post(url, json=payload)
        self.assertEqual(204, response.status_code)

        response = self.api_session.get(url)
        data = response.json()
        sub = [x for x in data["items"] if x["text"]["data"] == SUBTEXT][0]
        self.assertEqual(parent_id, sub["in_reply_to"])

    def test_update(self):
        url = f"{self.doc.absolute_url()}/@comments"
        OLD_TEXT = "comment 1"
        NEW_TEXT = "new text"

        self.api_session.post(url, json={"text": OLD_TEXT})

        response = self.api_session.get(url)
        data = response.json()
        item_texts = [x["text"]["data"] for x in data["items"]]
        self.assertNotIn(NEW_TEXT, item_texts)
        self.assertIn(OLD_TEXT, item_texts)
        comment = data["items"][0]

        payload = {"text": NEW_TEXT}
        response = self.api_session.patch(comment["@id"], json=payload)
        self.assertEqual(204, response.status_code)
        self.assertIn("location", response.headers)

        response = self.api_session.get(url)
        data = response.json()
        item_texts = [x["text"]["data"] for x in data["items"]]
        self.assertIn(NEW_TEXT, item_texts)
        self.assertNotIn(OLD_TEXT, item_texts)

    def test_permissions_delete_comment(self):
        url = f"{self.doc.absolute_url()}/@comments"

        response = self.api_session.post(url, json={"text": "comment"})
        self.assertEqual(204, response.status_code)

        response = self.api_session.get(url)
        comment_url = response.json()["items"][0]["@id"]
        self.assertFalse(comment_url.endswith("@comments"))
        self.assertTrue(response.json()["items"][0]["is_deletable"])

        # Other user may not delete this
        response = self.user_session.delete(comment_url)
        self.assertEqual(401, response.status_code)

        response = self.user_session.get(url)
        self.assertFalse(response.json()["items"][0]["is_deletable"])

        # The owner may
        response = self.api_session.delete(comment_url)
        self.assertEqual(204, response.status_code)

    def test_permissions_update_comment(self):
        url = f"{self.doc.absolute_url()}/@comments"

        response = self.api_session.post(url, json={"text": "comment"})
        self.assertEqual(204, response.status_code)

        response = self.api_session.get(url)
        comment_url = response.json()["items"][0]["@id"]
        self.assertFalse(comment_url.endswith("@comments"))
        self.assertTrue(response.json()["items"][0]["is_editable"])

        # Other user may not update this
        response = self.user_session.patch(comment_url, json={"text": "new"})
        self.assertEqual(401, response.status_code)

        response = self.user_session.get(url)
        self.assertFalse(response.json()["items"][0]["is_editable"])

        # The owner may
        response = self.api_session.patch(comment_url, json={"text": "new"})
        self.assertEqual(204, response.status_code)



from plone.app.multilingual.interfaces import ITranslatable
from plone.app.multilingual.interfaces import ITranslationManager
from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import IExpandableElement
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import ILanguage
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import Interface

import plone.protect.interfaces
import transaction


@implementer(IExpandableElement)
@adapter(ITranslatable, Interface)
class Translations:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, expand=False):
        result = {
            "translations": {"@id": f"{self.context.absolute_url()}/@translations"}
        }
        if not expand:
            return result

        translations = []
        manager = ITranslationManager(self.context)
        for language, translation in manager.get_restricted_translations().items():
            if language != ILanguage(self.context).get_language():
                translations.append(
                    {"@id": translation.absolute_url(), "language": language}
                )

        portal_state = getMultiAdapter(
            (self.context, self.request), name="plone_portal_state"
        )
        current_lang_nav_root = portal_state.navigation_root()

        if IPloneSiteRoot.providedBy(current_lang_nav_root):
            # We are not inside a LRF, bail off
            return result

        nav_root_manager = ITranslationManager(current_lang_nav_root)
        nav_root_translations = {}
        for (
            language,
            translation,
        ) in nav_root_manager.get_restricted_translations().items():
            nav_root_translations[language] = translation.absolute_url()

        result["translations"]["items"] = translations
        result["translations"]["root"] = nav_root_translations
        return result


class TranslationInfo(Service):
    """Get translation information"""

    def reply(self):
        translations = Translations(self.context, self.request)
        return translations(expand=True)["translations"]


class LinkTranslations(Service):
    """Link two content objects as translations of each other"""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.portal = getMultiAdapter(
            (self.context, self.request), name="plone_portal_state"
        ).portal()
        self.portal_url = self.portal.absolute_url()
        self.catalog = getToolByName(self.context, "portal_catalog")

    def reply(self):
        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        data = json_body(self.request)
        id_ = data.get("id", None)
        if id_ is None:
            self.request.response.setStatus(400)
            return dict(
                error=dict(type="BadRequest", message="Missing content id to link to")
            )

        target = self.get_object(id_)
        if target is None:
            self.request.response.setStatus(400)
            return dict(error=dict(type="BadRequest", message="Content does not exist"))
        elif target.portal_type == "LRF":
            self.request.response.setStatus(400)
            return dict(
                error=dict(
                    type="BadRequest",
                    message="Language Root Folders can only be linked between each other",
                )
            )

        target_language = ILanguage(target).get_language()
        manager = ITranslationManager(self.context)
        current_translation = manager.get_translation(target_language)
        target_manager = ITranslationManager(target)
        target_translation = target_manager.get_translation(self.context.language)
        if current_translation is not None:
            self.request.response.setStatus(400)
            return dict(
                error=dict(
                    type="BadRequest",
                    message="Source already translated into language {}".format(
                        target_language
                    ),
                )
            )
        if target_translation is not None:
            self.request.response.setStatus(400)
            return dict(
                error=dict(
                    type="BadRequest",
                    message="Target already translated into language {}".format(
                        target_language
                    ),
                )
            )

        manager.register_translation(target_language, target)
        # We want to leave a log in the transaction that the link has been executed
        ts = transaction.get()
        ts.note(
            f'Linked translation {"/".join(self.context.getPhysicalPath())} ({self.context.language}) -> {"/".join(target.getPhysicalPath())} ({target_language})'
        )

        self.request.response.setStatus(201)
        self.request.response.setHeader("Location", self.context.absolute_url())
        return {}

    def get_object(self, key):
        if key.startswith(self.portal_url):
            # Resolve by URL
            key = key[len(self.portal_url) + 1 :]
            return self.portal.restrictedTraverse(key, None)
        elif key.startswith("/"):
            # Resolve by path
            return self.portal.restrictedTraverse(key.lstrip("/"), None)
        else:
            # Resolve by UID
            brain = self.catalog(UID=key)
            if brain:
                return brain[0].getObject()


class UnlinkTranslations(Service):
    """Unlink the translations for a content object"""

    def reply(self):
        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        data = json_body(self.request)
        manager = ITranslationManager(self.context)
        language = data.get("language", None)
        if language is None:
            self.request.response.setStatus(400)
            return dict(
                error=dict(
                    type="BadRequest",
                    message="You need to provide the language to unlink",
                )
            )
        elif language not in list(manager.get_translations()):
            self.request.response.setStatus(400)
            return dict(
                error=dict(
                    type="BadRequest",
                    message=f"This objects is not translated into {language}",
                )
            )
        elif self.context.portal_type == "LRF":
            self.request.response.setStatus(400)
            return dict(
                error=dict(
                    type="BadRequest",
                    message="Language Root Folders cannot be unlinked",
                )
            )

        manager.remove_translation(language)
        # We want to leave a log in the transaction that the unlink has been executed
        ts = transaction.get()
        ts.note(
            f'Unlinked translation for {language} in {"/".join(self.context.getPhysicalPath())} ({self.context.language})'
        )

        return self.reply_no_content()



from plone import api
from plone.app.testing import login
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.dexterity.utils import createContentInContainer
from plone.app.multilingual.interfaces import IPloneAppMultilingualInstalled
from plone.app.multilingual.interfaces import ITranslationManager
from plone.restapi.testing import PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING
from plone.restapi.testing import PLONE_RESTAPI_DX_PAM_INTEGRATION_TESTING
from Products.CMFPlone.interfaces import ILanguage

from zope.component import getMultiAdapter
from zope.interface import alsoProvides

import requests
import transaction
import unittest


class TestTranslationInfo(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_PAM_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        login(self.portal, SITE_OWNER_NAME)
        self.en_content = createContentInContainer(
            self.portal["en"], "Document", title="Test document"
        )
        self.es_content = createContentInContainer(
            self.portal["es"], "Document", title="Test document"
        )
        ITranslationManager(self.en_content).register_translation("es", self.es_content)

    def test_translation_info_includes_translations(self):
        tinfo = getMultiAdapter(
            (self.en_content, self.request), name="GET_application_json_@translations"
        )

        info = tinfo.reply()
        self.assertIn("items", info)
        self.assertEqual(1, len(info["items"]))

    def test_correct_translation_information(self):
        tinfo = getMultiAdapter(
            (self.en_content, self.request), name="GET_application_json_@translations"
        )

        info = tinfo.reply()
        tinfo_es = info["items"][0]
        self.assertEqual(self.es_content.absolute_url(), tinfo_es["@id"])
        self.assertEqual(
            ILanguage(self.es_content).get_language(), tinfo_es["language"]
        )

    def test_translation_info_includes_root_translations(self):
        tinfo = getMultiAdapter(
            (self.en_content, self.request), name="GET_application_json_@translations"
        )

        info = tinfo.reply()
        self.assertIn("root", info)
        self.assertEqual(4, len(info["root"]))


class TestLinkContentsAsTranslations(unittest.TestCase):
    layer = PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        login(self.portal, SITE_OWNER_NAME)
        self.en_content = createContentInContainer(
            self.portal["en"], "Document", title="Test document"
        )
        self.es_content = createContentInContainer(
            self.portal["es"], "Document", title="Test document"
        )
        transaction.commit()

    def test_translation_linking_by_url(self):
        response = requests.post(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": self.es_content.absolute_url()},
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()
        manager = ITranslationManager(self.en_content)
        for language, translation in manager.get_translations():
            if language == ILanguage(self.es_content).get_language():
                self.assertEqual(translation, self.es_content)

    def test_translation_linking_by_path(self):
        response = requests.post(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": "/es/test-document"},
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()
        manager = ITranslationManager(self.en_content)
        for language, translation in manager.get_translations():
            if language == ILanguage(self.es_content).get_language():
                self.assertEqual(translation, self.es_content)

    def test_translation_linking_by_uid(self):
        response = requests.post(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": self.es_content.UID()},
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()
        manager = ITranslationManager(self.en_content)
        for language, translation in manager.get_translations():
            if language == ILanguage(self.es_content).get_language():
                self.assertEqual(translation, self.es_content)

    def test_calling_endpoint_without_id_gives_400(self):
        response = requests.post(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={},
        )
        self.assertEqual(400, response.status_code)

    def test_calling_with_an_already_translated_content_gives_400(self):
        ITranslationManager(self.en_content).register_translation("es", self.es_content)
        transaction.commit()
        response = requests.post(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": self.es_content.absolute_url()},
        )
        self.assertEqual(400, response.status_code)

    def test_calling_with_inexistent_content_gives_400(self):
        response = requests.post(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": "http://this-content-does-not-exist"},
        )
        self.assertEqual(400, response.status_code)

    def test_get_translations_on_content_with_no_permissions(self):
        response = requests.post(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": self.es_content.absolute_url()},
        )
        self.assertEqual(201, response.status_code)
        api.content.transition(self.en_content, "publish")
        transaction.commit()

        response = requests.get(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
        )

        self.assertEqual(200, response.status_code)
        response = response.json()
        self.assertTrue(len(response["items"]) == 0)

    def test_link_translation_with_an_already_translated_content_returns_400(self):
        ITranslationManager(self.en_content).register_translation("es", self.es_content)
        transaction.commit()
        response = requests.post(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": self.es_content.absolute_url()},
        )
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            response.json()["error"]["message"],
            "Source already translated into language es",
        )

    def test_link_translation_with_target_already_linked_to_other_object_returns_400(
        self,
    ):
        self.en_content_2 = createContentInContainer(
            self.portal["en"], "Document", title="Test document 2"
        )
        ITranslationManager(self.en_content_2).register_translation(
            "es", self.es_content
        )
        transaction.commit()
        response = requests.post(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": self.es_content.absolute_url()},
        )
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            response.json()["error"]["message"],
            "Target already translated into language es",
        )

    def test_link_translation_with_LFRs_not_possible_since_they_are_protected_returns_400(
        self,
    ):
        response = requests.post(
            f"{self.es_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"id": self.portal["en"].absolute_url()},
        )
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            response.json()["error"]["message"],
            "Language Root Folders can only be linked between each other",
        )


class TestUnLinkContentTranslations(unittest.TestCase):
    layer = PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        login(self.portal, SITE_OWNER_NAME)
        self.en_content = createContentInContainer(
            self.portal["en"], "Document", title="Test document"
        )
        self.es_content = createContentInContainer(
            self.portal["es"], "Document", title="Test document"
        )
        ITranslationManager(self.en_content).register_translation("es", self.es_content)
        transaction.commit()

    def test_translation_unlinking_succeeds(self):
        response = requests.delete(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"language": "es"},
        )
        self.assertEqual(204, response.status_code)
        transaction.begin()
        manager = ITranslationManager(self.en_content)
        self.assertNotIn(
            ILanguage(self.es_content).get_language(), list(manager.get_translations())
        )

    def test_calling_endpoint_without_language_gives_400(self):
        response = requests.delete(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={},
        )
        self.assertEqual(400, response.status_code)

    def test_calling_with_an_untranslated_content_gives_400(self):
        ITranslationManager(self.en_content).remove_translation("es")
        transaction.commit()
        response = requests.delete(
            f"{self.en_content.absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"language": "es"},
        )
        self.assertEqual(400, response.status_code)

    def test_translation_unlinking_a_LRF_errors(self):
        response = requests.delete(
            f"{self.portal['en'].absolute_url()}/@translations",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"language": "es"},
        )
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            response.json()["error"]["message"],
            "Language Root Folders cannot be unlinked",
        )


class TestCreateContentsAsTranslations(unittest.TestCase):
    layer = PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        login(self.portal, SITE_OWNER_NAME)
        self.es_content = createContentInContainer(
            self.portal["es"], "Document", title="Test document"
        )
        transaction.commit()

    def test_post_to_folder_creates_document_translated(self):
        response = requests.post(
            f"{self.portal.absolute_url()}/de",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "Document",
                "id": "mydocument",
                "title": "My Document DE",
                "translation_of": self.es_content.UID(),
                "language": "de",
            },
        )
        self.assertEqual(201, response.status_code)
        transaction.commit()

        manager = ITranslationManager(self.es_content)

        self.assertTrue("de" in manager.get_translations())
        self.assertEqual("My Document DE", manager.get_translations()["de"].title)

        self.assertEqual("Document", response.json().get("@type"))
        self.assertEqual("mydocument", response.json().get("id"))
        self.assertEqual("My Document DE", response.json().get("title"))


class TestTranslationLocator(unittest.TestCase):
    layer = PLONE_RESTAPI_DX_PAM_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        self.request = self.layer["request"]
        alsoProvides(self.layer["request"], IPloneAppMultilingualInstalled)
        login(self.portal, SITE_OWNER_NAME)
        self.es_content = createContentInContainer(
            self.portal["es"], "Document", title="Test document"
        )
        transaction.commit()

    def test_translation_locator(self):
        response = requests.get(
            "{}/@translation-locator?target_language=de".format(
                self.es_content.absolute_url()
            ),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )
        self.assertEqual(200, response.status_code)

        self.assertEqual(self.portal_url + "/de", response.json().get("@id"))



from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.globalrequest import getRequest


def get_scales(context, field, width, height):
    """Get a dictionary of available scales for a particular image field,
    with the actual dimensions (aspect ratio of the original image).
    """
    scales = {}
    request = getRequest()
    images_view = getMultiAdapter((context, request), name="images")

    for name, actual_width, actual_height in get_scale_infos():
        # Try first with scale name
        scale = images_view.scale(field.__name__, scale=name)
        if scale is None:
            # Sometimes it fails, but we can create it
            # using scale sizes
            scale = images_view.scale(
                field.__name__, width=actual_width, height=actual_height
            )

        if scale is None:
            # If we still can't get a scale, it's probably a corrupt image
            continue

        url = scale.url
        actual_width = scale.width
        actual_height = scale.height

        scales[name] = {
            "download": url,
            "width": actual_width,
            "height": actual_height,
        }

    return scales


def get_original_image_url(context, fieldname, width, height):
    request = getRequest()
    images_view = getMultiAdapter((context, request), name="images")
    scale = images_view.scale(
        fieldname, width=width, height=height, direction="thumbnail"
    )
    if scale:
        return scale.url
    # Corrupt images may not have a scale.


def get_actual_scale(dimensions, bbox):
    """Given dimensions of an original image, and a bounding box of a scale,
    calculates what actual dimensions that scaled image would have,
    maintaining aspect ratio.

    This is supposed to emulate / predict the behavior of Plone's
    ImageScaling implementations.
    """
    width, height = dimensions
    max_width, max_height = bbox
    resize_ratio = min(max_width / width, max_height / height)

    # Plone doesn't upscale images for the default named scales - limit
    # to actual image dimensions
    resize_ratio = min(resize_ratio, 1.0)

    scaled_dimensions = int(width * resize_ratio), int(height * resize_ratio)

    # Don't produce zero pixel lengths
    scaled_dimensions = tuple(max(1, dim) for dim in scaled_dimensions)
    return scaled_dimensions


def get_scale_infos():
    """Returns a list of (name, width, height) 3-tuples of the
    available image scales.
    """
    from plone.registry.interfaces import IRegistry

    registry = getUtility(IRegistry)
    from Products.CMFPlone.interfaces import IImagingSchema

    imaging_settings = registry.forInterface(IImagingSchema, prefix="plone")
    allowed_sizes = imaging_settings.allowed_sizes

    def split_scale_info(allowed_size):
        name, dims = allowed_size.split(" ")
        width, height = list(map(int, dims.split(":")))
        return name, width, height

    return [split_scale_info(size) for size in allowed_sizes]



from plone.restapi.interfaces import IExpandableElement
from plone.restapi.services import Service
from Products.CMFCore.interfaces import IActionCategory
from Products.CMFCore.utils import getToolByName
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import Interface


@implementer(IExpandableElement)
@adapter(Interface, Interface)
class Actions:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, expand=False):
        result = {"actions": {"@id": f"{self.context.absolute_url()}/@actions"}}
        if not expand:
            return result

        context_state = getMultiAdapter(
            (self.context, self.request), name="plone_context_state"
        )

        categories = self.request.form.get("categories", self.all_categories)
        data = {}
        for category in categories:
            category_action_data = []
            actions = context_state.actions(category=category)
            for action in actions:
                category_action_data.append(
                    {
                        "title": translate(action["title"], context=self.request),
                        "id": action["id"],
                        "icon": action["icon"],
                    }
                )
            data[category] = category_action_data
        return {"actions": data}

    @property
    def all_categories(self):
        portal_actions = getToolByName(self.context, "portal_actions")
        categories = []
        for id, obj in portal_actions.objectItems():
            if IActionCategory.providedBy(obj):
                categories.append(id)
        return categories


class ActionsGet(Service):
    def reply(self):
        actions = Actions(self.context, self.request)
        return actions(expand=True)["actions"]



from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.dexterity.utils import createContentInContainer
from plone.restapi.batching import DEFAULT_BATCH_SIZE
from plone.restapi.batching import HypermediaBatch
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.restapi.testing import RelativeSession
from plone.restapi.tests.helpers import result_paths
from urllib.parse import parse_qsl
from urllib.parse import urlparse

import transaction
import unittest


class TestBatchingDXBase(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        self.request = self.portal.REQUEST

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

    def tearDown(self):
        self.api_session.close()

    def _create_doc(self, container, number):
        createContentInContainer(
            container,
            "DXTestDocument",
            id="doc-%s" % str(number + 1),
            title="Document %s" % str(number + 1),
        )


class TestBatchingSearch(TestBatchingDXBase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        super().setUp()

        folder = createContentInContainer(self.portal, "Folder", id="folder")

        for i in range(5):
            self._create_doc(folder, i)
        transaction.commit()

    def test_contains_canonical_url(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/folder/@search?b_start=2&b_size=2")

        # Response should contain canonical URL without batching params
        self.assertEqual(response.json()["@id"], self.portal_url + "/folder/@search")

    def test_canonical_url_preserves_multiple_metadata_fields(self):
        qs = "b_start=2&b_size=2&metadata_fields=one&metadata_fields=two"
        response = self.api_session.get("/folder/@search?%s" % qs)

        # Response should contain canonical URL without batching params.
        # Argument lists like metadata_fields (same query string parameter
        # repeated multiple times) should be preserved.

        original_qs = parse_qsl(qs)
        canonicalized_qs = parse_qsl(urlparse(response.json()["@id"]).query)

        self.assertEqual(
            set(original_qs) - {("b_size", "2"), ("b_start", "2")},
            set(canonicalized_qs),
        )

    def test_contains_batching_links(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/folder/@search?b_start=2&b_size=2")

        # Batch info in response should contain appropriate batching links
        batch_info = response.json()["batching"]

        self.assertDictEqual(
            {
                "@id": self.portal_url + "/folder/@search?b_start=2&b_size=2",
                "first": self.portal_url + "/folder/@search?b_start=0&b_size=2",
                "next": self.portal_url + "/folder/@search?b_start=4&b_size=2",
                "prev": self.portal_url + "/folder/@search?b_start=0&b_size=2",
                "last": self.portal_url + "/folder/@search?b_start=4&b_size=2",
            },
            batch_info,
        )

    def test_contains_correct_batch_of_items(self):
        # Fetch the second page of the batch
        response = self.api_session.get(
            "/folder/@search?sort_on=path&b_start=2&b_size=2"
        )

        # Response should contain second batch of items
        self.assertEqual(
            ["/plone/folder/doc-2", "/plone/folder/doc-3"],
            result_paths(response.json()),
        )

    def test_total_item_count_is_correct(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/folder/@search?b_start=2&b_size=2")

        # Total count of items should be in items_total
        self.assertEqual(6, response.json()["items_total"])


class TestBatchingCollections(TestBatchingDXBase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        super().setUp()

        folder = createContentInContainer(self.portal, "Folder", id="folder")

        for i in range(5):
            self._create_doc(folder, i)

        collection = createContentInContainer(
            self.portal, "Collection", id="collection"
        )
        collection.query = [
            {
                "i": "path",
                "o": "plone.app.querystring.operation.string.path",
                "v": "/plone/folder/",
            }
        ]
        transaction.commit()

    def test_contains_canonical_url(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/collection?b_start=2&b_size=2")

        # Response should contain canonical URL without batching params
        self.assertEqual(response.json()["@id"], self.portal_url + "/collection")

    def test_contains_batching_links(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/collection?b_start=2&b_size=2")

        # Batch info in response should contain appropriate batching links
        batch_info = response.json()["batching"]

        self.assertDictEqual(
            {
                "@id": self.portal_url + "/collection?b_start=2&b_size=2",
                "first": self.portal_url + "/collection?b_start=0&b_size=2",
                "next": self.portal_url + "/collection?b_start=4&b_size=2",
                "prev": self.portal_url + "/collection?b_start=0&b_size=2",
                "last": self.portal_url + "/collection?b_start=4&b_size=2",
            },
            batch_info,
        )

    def test_contains_correct_batch_of_items(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/collection?b_start=2&b_size=2")

        # Response should contain second batch of items
        _result_paths = result_paths(response.json())
        self.assertEqual(2, len(_result_paths))
        self.assertTrue(
            all(path.startswith("/plone/folder/doc-") for path in _result_paths)
        )

    def test_total_item_count_is_correct(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/collection?b_start=2&b_size=2")

        # Total count of items should be in items_total
        self.assertEqual(6, response.json()["items_total"])

    def test_batching_links_omitted_if_resulset_fits_in_single_batch(self):
        response = self.api_session.get("/collection?b_size=100")
        self.assertNotIn("batching", list(response.json()))


class TestBatchingDXFolders(TestBatchingDXBase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        super().setUp()

        folder = createContentInContainer(self.portal, "Folder", id="folder")

        for i in range(5):
            self._create_doc(folder, i)
        transaction.commit()

    def test_contains_canonical_url(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/folder?b_start=2&b_size=2")

        # Response should contain canonical URL without batching params
        self.assertEqual(response.json()["@id"], self.portal_url + "/folder")

    def test_contains_batching_links(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/folder?b_start=2&b_size=2")

        # Batch info in response should contain appropriate batching links
        batch_info = response.json()["batching"]

        self.assertDictEqual(
            {
                "@id": self.portal_url + "/folder?b_start=2&b_size=2",
                "first": self.portal_url + "/folder?b_start=0&b_size=2",
                "next": self.portal_url + "/folder?b_start=4&b_size=2",
                "prev": self.portal_url + "/folder?b_start=0&b_size=2",
                "last": self.portal_url + "/folder?b_start=4&b_size=2",
            },
            batch_info,
        )

    def test_contains_correct_batch_of_items(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/folder?b_start=2&b_size=2")

        # Response should contain second batch of items
        _result_paths = result_paths(response.json())
        self.assertEqual(2, len(_result_paths))
        self.assertTrue(
            all(path.startswith("/plone/folder/doc-") for path in _result_paths)
        )

    def test_total_item_count_is_correct(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/folder?b_start=2&b_size=2")

        # Total count of items should be in items_total
        self.assertEqual(5, response.json()["items_total"])

    def test_batching_links_omitted_if_resulset_fits_in_single_batch(self):
        response = self.api_session.get("/folder?b_size=100")
        self.assertNotIn("batching", list(response.json()))

    def test_contains_batching_links_using_fullobjects(self):
        # Fetch the second page of the batch using fullobjects
        response = self.api_session.get("/folder?b_start=2&b_size=2&fullobjects")

        # Batch info in response should contain appropriate batching links
        batch_info = response.json()["batching"]

        self.assertDictEqual(
            {
                "@id": self.portal_url + "/folder?b_start=2&b_size=2&fullobjects",
                "first": self.portal_url + "/folder?b_start=0&b_size=2&fullobjects=",
                "next": self.portal_url + "/folder?b_start=4&b_size=2&fullobjects=",
                "prev": self.portal_url + "/folder?b_start=0&b_size=2&fullobjects=",
                "last": self.portal_url + "/folder?b_start=4&b_size=2&fullobjects=",
            },
            batch_info,
        )


class TestBatchingSiteRoot(TestBatchingDXBase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        super().setUp()

        for i in range(5):
            self._create_doc(self.portal, i)
        transaction.commit()

    def test_contains_canonical_url(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/?b_start=2&b_size=2")

        # Response should contain canonical URL without batching params
        self.assertEqual(response.json()["@id"], self.portal_url)

    def test_contains_batching_links(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/?b_start=2&b_size=2")

        # Batch info in response should contain appropriate batching links
        batch_info = response.json()["batching"]

        self.assertDictEqual(
            {
                "@id": self.portal_url + "/?b_start=2&b_size=2",
                "first": self.portal_url + "/?b_start=0&b_size=2",
                "next": self.portal_url + "/?b_start=4&b_size=2",
                "prev": self.portal_url + "/?b_start=0&b_size=2",
                "last": self.portal_url + "/?b_start=4&b_size=2",
            },
            batch_info,
        )

    def test_contains_correct_batch_of_items(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/?b_start=2&b_size=2")

        # Response should contain second batch of items
        _result_paths = result_paths(response.json())
        self.assertEqual(2, len(_result_paths))
        self.assertTrue(all(path.startswith("/plone/doc-") for path in _result_paths))

    def test_total_item_count_is_correct(self):
        # Fetch the second page of the batch
        response = self.api_session.get("/?b_start=2&b_size=2")

        # Total count of items should be in items_total
        self.assertEqual(5, response.json()["items_total"])

    def test_batching_links_omitted_if_resulset_fits_in_single_batch(self):
        response = self.api_session.get("/folder?b_size=100")
        self.assertNotIn("batching", list(response.json()))


class TestHypermediaBatch(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.portal.REQUEST

    def test_items_total(self):
        items = list(range(1, 26))
        self.request.form["b_size"] = 10
        batch = HypermediaBatch(self.request, items)
        # items_total should be total number of items in the sequence
        self.assertEqual(25, batch.items_total)

    def test_default_batch_size(self):
        items = list(range(1, 27))
        batch = HypermediaBatch(self.request, items)
        self.assertEqual(DEFAULT_BATCH_SIZE, len(list(batch)))

    def test_custom_batch_size(self):
        items = list(range(1, 26))
        self.request.form["b_size"] = 5
        batch = HypermediaBatch(self.request, items)
        # Batch size should be customizable via request
        self.assertEqual(5, len(list(batch)))

    def test_default_batch_start(self):
        items = list(range(1, 26))
        self.request.form["b_size"] = 10
        batch = HypermediaBatch(self.request, items)
        # Batch should start on first item by default
        self.assertEqual(list(range(1, 11)), list(batch))

    def test_custom_batch_start(self):
        items = list(range(1, 26))
        self.request.form["b_size"] = 10
        self.request.form["b_start"] = 5
        batch = HypermediaBatch(self.request, items)
        # Batch start should be customizable via request
        self.assertEqual(list(range(6, 16)), list(batch))

    def test_custom_start_and_size_can_be_combined(self):
        items = list(range(1, 26))
        self.request.form["b_size"] = 5
        self.request.form["b_start"] = 5
        batch = HypermediaBatch(self.request, items)
        # Should be able to combine custom batch start and size
        self.assertListEqual(list(range(6, 11)), list(batch))

    def test_canonical_url(self):
        items = list(range(1, 26))
        self.request.form["b_size"] = 10
        batch = HypermediaBatch(self.request, items)
        self.assertEqual("http://nohost", batch.canonical_url)

    def test_canonical_url_preserves_query_string_params(self):
        items = list(range(1, 26))

        self.request.form["b_size"] = 10
        self.request["QUERY_STRING"] = "one=1&two=2"
        batch = HypermediaBatch(self.request, items)

        parsed_url = urlparse(batch.canonical_url)
        qs_params = dict(parse_qsl(parsed_url.query))

        self.assertEqual({"one": "1", "two": "2"}, qs_params)
        self.assertEqual("nohost", parsed_url.netloc)
        self.assertEqual("", parsed_url.path)

    def test_canonical_url_preserves_list_like_query_string_params(self):
        items = list(range(1, 26))

        self.request.form["b_size"] = 10
        self.request["QUERY_STRING"] = "foolist=1&foolist=2"
        batch = HypermediaBatch(self.request, items)

        # Argument lists (same query string parameter repeated multiple
        # times) should be preserved.

        self.assertEqual(
            {("foolist", "1"), ("foolist", "2")},
            set(parse_qsl(urlparse(batch.canonical_url).query)),
        )

    def test_canonical_url_strips_batching_params(self):
        items = list(range(1, 26))

        self.request.form["b_size"] = 10
        self.request["QUERY_STRING"] = "one=1&b_size=10&b_start=20&two=2"
        batch = HypermediaBatch(self.request, items)

        parsed_url = urlparse(batch.canonical_url)
        qs_params = dict(parse_qsl(parsed_url.query))

        self.assertEqual({"one": "1", "two": "2"}, qs_params)
        self.assertEqual("nohost", parsed_url.netloc)
        self.assertEqual("", parsed_url.path)

    def test_canonical_url_strips_sorting_params(self):
        items = list(range(1, 26))

        self.request["QUERY_STRING"] = "one=1&sort_on=path&two=2"
        batch = HypermediaBatch(self.request, items)

        parsed_url = urlparse(batch.canonical_url)
        qs_params = dict(parse_qsl(parsed_url.query))

        self.assertEqual({"one": "1", "two": "2"}, qs_params)
        self.assertEqual("nohost", parsed_url.netloc)
        self.assertEqual("", parsed_url.path)

    def test_current_batch_url(self):
        items = list(range(1, 26))

        self.request.form["b_size"] = 10
        self.request["ACTUAL_URL"] = "http://nohost"
        self.request["QUERY_STRING"] = "b_size=10&b_start=20"
        batch = HypermediaBatch(self.request, items)
        self.assertEqual("http://nohost?b_size=10&b_start=20", batch.current_batch_url)

    def test_batching_links_omitted_if_resultset_fits_in_single_batch(self):
        items = list(range(1, 5))

        self.request.form["b_size"] = 10
        batch = HypermediaBatch(self.request, items)
        self.assertEqual(None, batch.links)

    def test_first_link_contained(self):
        items = list(range(1, 26))

        self.request.form["b_size"] = 10
        batch = HypermediaBatch(self.request, items)
        self.assertDictContainsSubset({"first": "http://nohost?b_start=0"}, batch.links)

    def test_first_link_preserves_list_like_querystring_params(self):
        items = list(range(1, 26))

        self.request.form["b_size"] = 10
        self.request["QUERY_STRING"] = "foolist=1&foolist=2"
        batch = HypermediaBatch(self.request, items)

        # Argument lists (same query string parameter repeated multiple
        # times) should be preserved.

        batch_params = {("b_start", "0"), ("b_size", "10")}
        self.assertEqual(
            {("foolist", "1"), ("foolist", "2")},
            set(parse_qsl(urlparse(batch.links["first"]).query)) - batch_params,
        )

    def test_last_link_contained(self):
        items = list(range(1, 26))

        self.request.form["b_size"] = 10
        batch = HypermediaBatch(self.request, items)
        self.assertDictContainsSubset({"last": "http://nohost?b_start=20"}, batch.links)

    def test_next_link_contained_if_necessary(self):
        items = list(range(1, 26))

        self.request.form["b_size"] = 10
        batch = HypermediaBatch(self.request, items)
        self.assertDictContainsSubset({"next": "http://nohost?b_start=10"}, batch.links)

    def test_next_link_omitted_on_last_page(self):
        items = list(range(1, 26))

        # Start on last page
        self.request.form["b_size"] = 10
        self.request.form["b_start"] = 20
        batch = HypermediaBatch(self.request, items)
        self.assertSetEqual({"@id", "first", "prev", "last"}, set(batch.links))

    def test_prev_link_contained_if_necessary(self):
        items = list(range(1, 26))

        # Start on third page
        self.request.form["b_size"] = 10
        self.request.form["b_start"] = 20
        batch = HypermediaBatch(self.request, items)
        self.assertDictContainsSubset({"prev": "http://nohost?b_start=10"}, batch.links)

    def test_prev_link_omitted_on_first_page(self):
        items = list(range(1, 26))

        self.request.form["b_size"] = 10
        batch = HypermediaBatch(self.request, items)
        self.assertSetEqual({"@id", "first", "next", "last"}, set(batch.links))

    def test_no_gaps_or_duplicates_between_pages(self):
        items = list(range(1, 26))
        items_from_all_batches = []

        size = 10
        self.request.form["b_size"] = size

        for pagenumber in range(3):
            self.request.form["b_start"] = pagenumber * size
            batch = HypermediaBatch(self.request, items)
            items_from_all_batches.extend(list(batch))

        self.assertEqual(items, items_from_all_batches)

    def test_batch_start_never_drops_below_zero(self):
        items = list(range(1, 26))

        # Start in the middle of what would otherwise be the first batch
        self.request.form["b_size"] = 10
        self.request.form["b_start"] = 5
        batch = HypermediaBatch(self.request, items)
        self.assertEqual("http://nohost?b_start=0", batch.links["prev"])



from plone.restapi.deserializer import json_body
from plone.restapi.services import Service
from zExceptions import BadRequest
from zope.component import getMultiAdapter
from zope.interface import alsoProvides

import plone


class EmailNotificationPost(Service):
    def reply(self):
        data = json_body(self.request)

        sender_from_address = data.get("from", None)
        message = data.get("message", None)
        sender_fullname = data.get("name", "")
        subject = data.get("subject", "")

        if not sender_from_address or not message:
            raise BadRequest("Missing from or message parameters")

        overview_controlpanel = getMultiAdapter(
            (self.context, self.request), name="overview-controlpanel"
        )
        if overview_controlpanel.mailhost_warning():
            raise BadRequest("MailHost is not configured.")

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        contact_info_view = getMultiAdapter(
            (self.context, self.request), name="contact-info"
        )

        contact_info_view.send_message(
            dict(
                message=message,
                subject=subject,
                sender_from_address=sender_from_address,
                sender_fullname=sender_fullname,
            )
        )

        return self.reply_no_content()



from plone.restapi.interfaces import IExpandableElement
from plone.restapi.services import Service
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.interface import Interface


@implementer(IExpandableElement)
@adapter(Interface, Interface)
class Breadcrumbs:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, expand=False):
        result = {"breadcrumbs": {"@id": f"{self.context.absolute_url()}/@breadcrumbs"}}
        if not expand:
            return result

        portal_state = getMultiAdapter(
            (self.context, self.request), name="plone_portal_state"
        )
        breadcrumbs_view = getMultiAdapter(
            (self.context, self.request), name="breadcrumbs_view"
        )
        items = []
        for crumb in breadcrumbs_view.breadcrumbs():
            item = {
                "title": crumb["Title"],
                "@id": crumb["absolute_url"],
            }
            if crumb.get("nav_title", False):
                item.update({"title": crumb["nav_title"]})

            items.append(item)

        result["breadcrumbs"]["items"] = items
        result["breadcrumbs"]["root"] = portal_state.navigation_root().absolute_url()
        return result


class BreadcrumbsGet(Service):
    def reply(self):
        breadcrumbs = Breadcrumbs(self.context, self.request)
        return breadcrumbs(expand=True)["breadcrumbs"]



from plone.dexterity.interfaces import IDexterityContent
from plone.registry.interfaces import IRegistry
from plone.restapi.controlpanels import IControlpanel
from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import IDeserializeFromJson
from plone.restapi.interfaces import IFieldDeserializer
from z3c.form.interfaces import IManagerValidator
from zExceptions import BadRequest
from zope.component import adapter
from zope.component import getUtility
from zope.component import queryMultiAdapter
from zope.interface import implementer
from zope.schema import getFields
from zope.schema.interfaces import ValidationError


@implementer(IDexterityContent)
class FakeDXContext:
    """Fake DX content class, so we can re-use the DX field deserializers"""


@implementer(IDeserializeFromJson)
@adapter(IControlpanel)
class ControlpanelDeserializeFromJson:
    def __init__(self, controlpanel):
        self.controlpanel = controlpanel
        self.schema = self.controlpanel.schema
        self.schema_prefix = self.controlpanel.schema_prefix

        self.registry = getUtility(IRegistry)

        self.context = self.controlpanel.context
        self.request = self.controlpanel.request

    def __call__(self):
        data = json_body(self.controlpanel.request)

        proxy = self.registry.forInterface(self.schema, prefix=self.schema_prefix)

        schema_data = {}
        errors = []

        # Make a fake context
        fake_context = FakeDXContext()

        for name, field in getFields(self.schema).items():
            field_data = schema_data.setdefault(self.schema, {})

            if field.readonly:
                continue

            if name in data:
                deserializer = queryMultiAdapter(
                    (field, fake_context, self.request), IFieldDeserializer
                )

                try:
                    # Make it sane
                    value = deserializer(data[name])
                    # Validate required etc
                    field.validate(value)
                    # Set the value.
                    setattr(proxy, name, value)
                except ValueError as e:
                    errors.append({"message": str(e), "field": name, "error": e})
                except ValidationError as e:
                    errors.append({"message": e.doc(), "field": name, "error": e})
                else:
                    field_data[name] = value

        # Validate schemata
        for schema, field_data in schema_data.items():
            validator = queryMultiAdapter(
                (self.context, self.request, None, schema, None), IManagerValidator
            )
            for error in validator.validate(field_data):
                errors.append({"error": error, "message": str(error)})

        if errors:
            raise BadRequest(errors)



from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class UsersDelete(Service):
    """Deletes a user."""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Consume any path segments after /@users as parameters
        self.params.append(name)
        return self

    @property
    def _get_user_id(self):
        if len(self.params) != 1:
            raise Exception("Must supply exactly one parameter (user id)")
        return self.params[0]

    def reply(self):
        portal = getSite()
        portal_membership = getToolByName(portal, "portal_membership")
        delete_successful = portal_membership.deleteMembers((self._get_user_id,))
        if delete_successful:
            return self.reply_no_content()
        else:
            return self.reply_no_content(status=404)



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession

import unittest


class TestQuerystringEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

    def tearDown(self):
        self.api_session.close()

    def test_endpoint_lists_indexes(self):
        response = self.api_session.get("/@querystring")

        self.assertEqual(response.status_code, 200)
        self.assertIn("indexes", response.json())
        self.assertIn("isFolderish", response.json()["indexes"])

    def test_endpoint_lists_sortable_indexes(self):
        response = self.api_session.get("/@querystring")

        self.assertEqual(response.status_code, 200)
        self.assertIn("sortable_indexes", response.json())
        self.assertIn("sortable_title", response.json()["sortable_indexes"])

    def test_endpoint_shows_field_config(self):
        response = self.api_session.get("/@querystring")

        self.assertEqual(response.status_code, 200)
        indexes = response.json()["indexes"]
        idx = indexes["Title"]

        expected_field_config = {
            "description": "Text search of an item's title",
            "enabled": True,
            "group": "Text",
            "operations": ["plone.app.querystring.operation.string.contains"],
            "operators": {
                "plone.app.querystring.operation.string.contains": {
                    "description": None,
                    "operation": "plone.app.querystring.queryparser._contains",
                    "title": "Contains",
                    "widget": "StringWidget",
                }
            },
            "sortable": False,
            "title": "Title",
            "values": {},
            "vocabulary": None,
        }
        # Be permissive with the check and only check the existing
        # attributes. (This gives plone.app.querystring to extend its schema
        # when that becomes necessary, while making sure that all code depending
        # on any existing attributes continues to work.)
        filtered_idx = {}
        for key in expected_field_config:
            filtered_idx[key] = idx.get(key, "NOT-FOUND")
        self.assertEqual(expected_field_config, filtered_idx)

    def test_endpoint_inlines_vocabularies(self):
        response = self.api_session.get("/@querystring")

        self.assertEqual(response.status_code, 200)
        indexes = response.json()["indexes"]
        idx = indexes["review_state"]

        self.assertDictContainsSubset(
            {
                "title": "Review state",
                "vocabulary": "plone.app.vocabularies.WorkflowStates",
            },
            idx,
        )

        expected_vocab_values = {
            "external": {"title": "Externally visible [external]"},
            "internal": {"title": "Internal draft [internal]"},
            "internally_published": {
                "title": "Internally published [internally_published]"
            },
            "pending": {"title": "Pending [pending]"},
            "private": {"title": "Private [private]"},
            "published": {"title": "Published with accent \xe9 [published]"},
            "visible": {"title": "Public draft [visible]"},
        }
        self.assertTrue(
            all(elem in idx["values"].items() for elem in expected_vocab_values.items())
        )

    def test_endpoint_inlines_operators(self):
        response = self.api_session.get("/@querystring")

        self.assertEqual(response.status_code, 200)
        indexes = response.json()["indexes"]
        idx = indexes["isDefaultPage"]

        self.assertDictContainsSubset(
            {
                "title": "Default Page",
                "operations": [
                    "plone.app.querystring.operation.boolean.isTrue",
                    "plone.app.querystring.operation.boolean.isFalse",
                ],
            },
            idx,
        )

        expected_operators = {
            "plone.app.querystring.operation.boolean.isFalse": {
                "description": None,
                "operation": "plone.app.querystring.queryparser._isFalse",
                "title": "No",
                "widget": None,
            },
            "plone.app.querystring.operation.boolean.isTrue": {
                "description": None,
                "operation": "plone.app.querystring.queryparser._isTrue",
                "title": "Yes",
                "widget": None,
            },
        }
        self.assertEqual(expected_operators, idx["operators"])

    def test_endpoint_includes_widgets_for_operators(self):
        response = self.api_session.get("/@querystring")

        self.assertEqual(response.status_code, 200)
        indexes = response.json()["indexes"]
        idx = indexes["getObjPositionInParent"]

        self.assertDictContainsSubset(
            {
                "title": "Order in folder",
                "operations": [
                    "plone.app.querystring.operation.int.is",
                    "plone.app.querystring.operation.int.lessThan",
                    "plone.app.querystring.operation.int.largerThan",
                ],
            },
            idx,
        )

        ops = idx["operators"]

        self.assertEqual(
            {
                "description": None,
                "operation": "plone.app.querystring.queryparser._intLargerThan",
                "title": "Larger than",
                "widget": "StringWidget",
            },
            ops["plone.app.querystring.operation.int.largerThan"],
        )



"""
$Id: base.py 1808 2007-02-06 11:39:11Z hazmat $
"""

from AccessControl import getSecurityManager
from DateTime import DateTime
from plone.app.iterate.interfaces import IBaseline
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.interfaces import keys
from plone.app.iterate.permissions import CheckoutPermission
from plone.memoize.instance import memoize
from plone.restapi.serializer.converters import json_compatible
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.log import logger
from Products.Five.browser import BrowserView


class WorkingCopyInfo(BrowserView):
    def __init__(self, context):
        self.context = context
        self.ref = None

    def get_working_copy_info(self):
        baseline = self.baseline()
        working_copy = self.working_copy()

        sm = getSecurityManager()

        # No Working Copy exists
        if baseline is None and working_copy is None:
            return (None, None)

        # Baseline is None (context is the baseline), and working copy exists
        if baseline is None and working_copy:
            self.ref = working_copy

            if (
                sm.checkPermission(ModifyPortalContent, self.context)
                or sm.checkPermission(CheckoutPermission, self.context)
                or sm.checkPermission(ModifyPortalContent, working_copy)
            ):
                return (
                    None,
                    {
                        "@id": working_copy.absolute_url(),
                        "title": working_copy.title,
                        "created": json_compatible(self.created()),
                        "creator_url": self.creator_url(),
                        "creator_name": self.creator_name(),
                    },
                )
            else:
                return (None, None)

        # Baseline exist (context is the working copy), and working copy exists
        if baseline and working_copy:
            self.ref = baseline

            if sm.checkPermission(
                ModifyPortalContent, self.context
            ) or sm.checkPermission(CheckoutPermission, baseline):
                return (
                    {"@id": baseline.absolute_url(), "title": baseline.title},
                    {
                        "@id": working_copy.absolute_url(),
                        "title": working_copy.title,
                        "created": json_compatible(self.created()),
                        "creator_url": self.creator_url(),
                        "creator_name": self.creator_name(),
                    },
                )
            else:
                return (None, None)

    @property
    @memoize
    def policy(self):
        return ICheckinCheckoutPolicy(self.context)

    @memoize
    def working_copy(self):
        return self.policy.getWorkingCopy()

    @memoize
    def baseline(self):
        return self.policy.getBaseline()

    @memoize
    def created(self):
        return self.properties.get(keys.checkout_time, DateTime())

    @memoize
    def creator(self):
        user_id = self.properties.get(keys.checkout_user)
        membership = getToolByName(self.context, "portal_membership")
        if not user_id:
            return membership.getAuthenticatedMember()
        return membership.getMemberById(user_id)

    @memoize
    def creator_url(self):
        creator = self.creator()
        if creator is not None:
            portal_url = getToolByName(self.context, "portal_url")
            return f"{portal_url()}/author/{creator.getId()}"

    @memoize
    def creator_name(self):
        creator = self.creator()
        if creator is not None:
            return creator.getProperty("fullname") or creator.getId()
        # User is not known by PAS. This may be due to LDAP issues, so we keep
        # the user and log this.
        name = self.properties.get(keys.checkout_user)
        if IBaseline.providedBy(self.context):
            warning_tpl = (
                "%s is a baseline of a plone.app.iterate checkout "
                'by an unknown user id "%s"'
            )
        else:
            # IWorkingCopy.providedBy(self.context)
            warning_tpl = (
                "%s is a working copy of a plone.app.iterate "
                'checkout by an unknown user id "%s"'
            )
        logger.warning(warning_tpl, self.context, name)
        return name

    @property
    @memoize
    def properties(self):
        if self.ref:
            return self.policy.getProperties(self.ref, default={})
        else:
            return {}



from datetime import date
from datetime import time
from datetime import timedelta
from DateTime import DateTime
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from plone.restapi.serializer.converters import json_compatible
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from unittest import TestCase
from z3c.relationfield.relation import RelationValue
from zope.component import getUtility
from zope.i18nmessageid import MessageFactory
from zope.intid.interfaces import IIntIds

import json
import Missing


class TestJsonCompatibleConverters(TestCase):
    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def test_None(self):
        self.assertEqual(None, json_compatible(None))
        self.assertEqual("null", json.dumps(json_compatible(None)))

    def test_error_when_type_unknown(self):
        err_regex = (
            r"^No converter for making <object object at [^>]*>"
            + r" \(<(class|type) \'object\'>\) JSON compatible.$"
        )

        with self.assertRaisesRegex(TypeError, err_regex):
            json_compatible(object())

        with self.assertRaisesRegex(TypeError, err_regex):
            json_compatible({"foo": [object()]})

    def test_True(self):
        self.assertEqual(True, json_compatible(True))
        self.assertEqual("true", json.dumps(json_compatible(True)))

    def test_False(self):
        self.assertEqual(False, json_compatible(False))
        self.assertEqual("false", json.dumps(json_compatible(False)))

    def test_unicode(self):
        self.assertEqual("foo", json_compatible("foo"))
        self.assertIsInstance(json_compatible("foo"), str)
        self.assertEqual('"foo"', json.dumps(json_compatible("foo")))

    def test_unicode_with_umlaut(self):
        self.assertEqual("Hall\xf6chen", json_compatible("Hall\xf6chen"))
        self.assertEqual(
            '"Hall\\u00f6chen"', json.dumps(json_compatible("Hall\xf6chen"))
        )

    def test_string_is_converted_to_unicode(self):
        # Standard library JSON works with unicode.
        self.assertEqual("foo", json_compatible("foo"))
        self.assertIsInstance(json_compatible("foo"), str)
        self.assertEqual('"foo"', json.dumps(json_compatible("foo")))

    def test_string_with_umlaut(self):
        # Standard library JSON works with unicode.
        self.assertEqual("Hall\xf6chen", json_compatible("Hallöchen"))
        self.assertEqual('"Hall\\u00f6chen"', json.dumps(json_compatible("Hallöchen")))

    def test_int(self):
        self.assertEqual(42, json_compatible(42))
        self.assertIsInstance(json_compatible(42), int)
        self.assertEqual("42", json.dumps(json_compatible(42)))

    def test_long(self):
        def _long(val):
            return int(val)

        self.assertEqual(_long(10), json_compatible(_long(10)))
        self.assertIsInstance(json_compatible(_long(10)), int)
        self.assertEqual("10", json.dumps(json_compatible(_long(10))))

    def test_float(self):
        self.assertEqual(1.4, json_compatible(1.4))
        self.assertIsInstance(json_compatible(1.4), float)
        self.assertEqual("1.4", json.dumps(json_compatible(1.4)))

    def test_list(self):
        self.assertEqual(["foo"], json_compatible(["foo"]))
        self.assertEqual('["foo"]', json.dumps(json_compatible(["foo"])))
        self.assertIsInstance(
            json_compatible(["foo"])[0],
            str,
            "List values should be converted recursively.",
        )

    def test_persistent_list(self):
        value = PersistentList(["foo"])
        self.assertEqual(["foo"], json_compatible(value))
        self.assertEqual('["foo"]', json.dumps(json_compatible(value)))
        self.assertIsInstance(
            json_compatible(value)[0],
            str,
            "PersistentList values should be converted" " recursively.",
        )

    def test_tuple(self):
        # Tuples are converted to list (json would do that anyway and
        # it is easier to implement it with map).
        self.assertEqual(["foo", None], json_compatible(("foo", None)))
        self.assertEqual('["foo"]', json.dumps(json_compatible(("foo",))))
        self.assertIsInstance(
            json_compatible(("foo",))[0],
            str,
            "Tuple values should be converted recursively.",
        )

    def test_frozenset(self):
        self.assertEqual(
            [[1, 1], [2, 2]], sorted(json_compatible(frozenset([(1, 1), (2, 2)])))
        )

    def test_set(self):
        self.assertEqual([[1, 1], [2, 2]], sorted(json_compatible({(1, 1), (2, 2)})))

    def test_dict(self):
        self.assertEqual(
            {"foo": True, "bar": None, "baz": 3},
            json_compatible({"foo": True, "bar": None, "baz": 3}),
        )
        self.assertEqual('{"foo": "bar"}', json.dumps(json_compatible({"foo": "bar"})))
        self.assertIsInstance(
            json_compatible(list({"foo": "bar"})[0]),
            str,
            "Dict keys should be converted recursively.",
        )
        self.assertIsInstance(
            json_compatible(list({"foo": "bar"}.values())[0]),
            str,
            "Dict values should be converted recursively.",
        )

    def test_dict_empty(self):
        self.assertEqual({}, json_compatible({}))
        self.assertEqual("{}", json.dumps(json_compatible({})))

    def test_persistent_mapping(self):
        value = PersistentMapping({"foo": "bar"})
        self.assertEqual({"foo": "bar"}, json_compatible(value))
        self.assertEqual('{"foo": "bar"}', json.dumps(json_compatible(value)))
        self.assertIsInstance(
            json_compatible(list(value)[0]),
            str,
            "Dict keys should be converted recursively.",
        )
        self.assertIsInstance(
            json_compatible(list(value.values())[0]),
            str,
            "Dict values should be converted recursively.",
        )

    def test_python_datetime(self):
        value = DateTime("2015/11/23 19:45:55.649027 GMT+3").asdatetime()
        self.assertEqual("2015-11-23T16:45:55+00:00", json_compatible(value))
        self.assertEqual(
            '"2015-11-23T16:45:55+00:00"', json.dumps(json_compatible(value))
        )

    def test_zope_DateTime(self):
        value = DateTime("2015/11/23 19:45:55.649027 GMT+3")
        self.assertEqual("2015-11-23T16:45:55+00:00", json_compatible(value))
        self.assertEqual(
            '"2015-11-23T16:45:55+00:00"', json.dumps(json_compatible(value))
        )

    def test_date(self):
        value = date(2015, 11, 23)
        self.assertEqual("2015-11-23", json_compatible(value))
        self.assertEqual('"2015-11-23"', json.dumps(json_compatible(value)))

    def test_time(self):
        value = time(19, 45, 55)
        self.assertEqual("19:45:55", json_compatible(value))
        self.assertEqual('"19:45:55"', json.dumps(json_compatible(value)))

    def test_timedelta(self):
        self.assertEqual(9.58, json_compatible(timedelta(seconds=9.58)))

    def test_broken_relation_value(self):
        self.assertEqual(None, json_compatible(RelationValue(12345)))

    def test_relation_value(self):
        portal = self.layer["portal"]
        doc1 = portal[
            portal.invokeFactory(
                "DXTestDocument",
                id="doc1",
                title="Document 1",
                description="Description",
            )
        ]
        intids = getUtility(IIntIds)
        self.assertEqual(
            {
                "@id": "http://nohost/plone/doc1",
                "@type": "DXTestDocument",
                "title": "Document 1",
                "description": "Description",
                "review_state": "private",
            },
            json_compatible(RelationValue(intids.getId(doc1))),
        )

    def test_i18n_message(self):
        _ = MessageFactory("plone.restapi.tests")
        msg = _("message_id", default="default message")
        self.assertEqual("default message", json_compatible(msg))

    def test_missing_value(self):
        self.assertEqual(None, json_compatible(Missing.Value))



from AccessControl import getSecurityManager
from Acquisition import aq_inner
from plone.app.discussion.interfaces import IDiscussionSettings
from plone.registry.interfaces import IRegistry
from zExceptions import Unauthorized
from zope.component import queryUtility
from zope.security.interfaces import IPermission


def edit_comment_allowed():
    # Check if editing comments is allowed in the registry
    registry = queryUtility(IRegistry)
    settings = registry.forInterface(IDiscussionSettings, check=False)
    return settings.edit_comment_enabled


def can_edit(comment):
    """Returns true if current user has the 'Edit comments'
    permission.
    """
    return bool(
        getSecurityManager().checkPermission("Edit comments", aq_inner(comment))
    )


def permission_exists(permission_id):
    permission = queryUtility(IPermission, permission_id)
    return permission is not None


def can_review(comment):
    """Returns true if current user has the 'Review comments' permission."""
    return bool(
        getSecurityManager().checkPermission("Review comments", aq_inner(comment))
    )


def can_delete(comment):
    """Returns true if current user has the 'Delete comments'
    permission.
    """
    if not permission_exists("plone.app.discussion.DeleteComments"):
        # Older versions of p.a.discussion do not support this yet.
        return can_review(comment)

    return bool(
        getSecurityManager().checkPermission("Delete comments", aq_inner(comment))
    )


def delete_own_comment_allowed():
    if not permission_exists("plone.app.discussion.DeleteOwnComments"):
        # Older versions of p.a.discussion do not support this yet.
        return False
    # Check if delete own comments is allowed in the registry
    registry = queryUtility(IRegistry)
    settings = registry.forInterface(IDiscussionSettings, check=False)
    return settings.delete_own_comment_enabled


def can_delete_own(comment):
    """Returns true if the current user could delete the comment if it had
    no replies. This is used to prepare hidden form buttons for JS.
    """
    if not permission_exists("plone.app.discussion.DeleteOwnComments"):
        # Older versions of p.a.discussion do not support this yet.
        return False
    try:
        return comment.restrictedTraverse("@@delete-own-comment").could_delete()
    except Unauthorized:
        return False



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.registry import field
from plone.registry.interfaces import IRegistry
from plone.registry.record import Record
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from zope.component import getUtility

import transaction
import unittest


class TestRegistry(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        registry = getUtility(IRegistry)
        record = Record(field.TextLine(title="Foo Bar"), "Lorem Ipsum")
        registry.records["foo.bar"] = record

        for counter in range(1, 100):
            record = Record(field.TextLine(title="Foo Bar"), "Lorem Ipsum")
            registry.records["foo.bar" + str(counter)] = record

        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_get_registry_record(self):
        response = self.api_session.get("/@registry/foo.bar")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), "Lorem Ipsum")

    def test_update_registry_record(self):
        registry = getUtility(IRegistry)
        payload = {"foo.bar": "lorem ipsum"}
        response = self.api_session.patch("/@registry", json=payload)
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        self.assertEqual(registry["foo.bar"], "lorem ipsum")

    def test_update_several_registry_records(self):
        registry = getUtility(IRegistry)
        record = Record(field.TextLine(title="Foo Bar Baz"), "Lorem Ipsum Dolor")
        registry.records["foo.bar.baz"] = record
        transaction.commit()
        payload = {"foo.bar": "lorem ipsum", "foo.bar.baz": "lorem ipsum dolor"}
        response = self.api_session.patch("/@registry", json=payload)
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        self.assertEqual(registry["foo.bar"], "lorem ipsum")
        self.assertEqual(registry["foo.bar.baz"], "lorem ipsum dolor")

    def test_update_non_existing_registry_record(self):
        payload = {"foo.bar.baz": "lorem ipsum"}
        response = self.api_session.patch("/@registry", json=payload)
        self.assertEqual(response.status_code, 500)

    def test_get_listing(self):
        response = self.api_session.get("/@registry")

        self.assertEqual(response.status_code, 200)
        response = response.json()
        self.assertIn("items", response)
        self.assertIn("batching", response)
        self.assertIn("next", response["batching"])



from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from Products.CMFCore.utils import getToolByName

import unittest


class TestDatabaseServiceFunctional(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        self.request = self.portal.REQUEST
        self.catalog = getToolByName(self.portal, "portal_catalog")

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

    def tearDown(self):
        self.api_session.close()

    def test_get_system(self):
        response = self.api_session.get("/@database")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get("Content-Type"), "application/json")

        results = response.json()
        self.assertEqual(results["@id"], self.portal.absolute_url() + "/@database")
        self.assertTrue("cache_length" in results.keys())
        self.assertTrue("cache_length_bytes" in results.keys())
        self.assertTrue("cache_detail_length" in results.keys())
        self.assertTrue("cache_size" in results.keys())
        self.assertTrue("database_size" in results.keys())
        self.assertTrue("db_name" in results.keys())
        self.assertTrue("db_size" in results.keys())



from datetime import datetime
from DateTime import DateTime
from datetime import timedelta
from operator import itemgetter
from plone import api
from plone.app.discussion.interfaces import IConversation
from plone.app.discussion.interfaces import IDiscussionSettings
from plone.app.discussion.interfaces import IReplies
from plone.app.layout.viewlets.content import ContentHistoryViewlet
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.locking.interfaces import ILockable
from plone.locking.interfaces import ITTWLockable
from plone.registry.interfaces import IRegistry
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.tests.statictime import StaticTime
from zope.component import createObject
from zope.component import getUtility
from zope.interface import alsoProvides
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.restapi.serializer.working_copy import WorkingCopyInfo
from plone.restapi.testing import PLONE_RESTAPI_ITERATE_FUNCTIONAL_TESTING
from plone.restapi.serializer.converters import json_compatible

import transaction
import unittest


class TestStaticTime(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()

        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        registry = getUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True

        transaction.commit()

    def create_document(self, id_):
        self.portal.invokeFactory("Document", id=id_)
        document = self.portal[id_]
        document.title = "My title"
        return document

    def create_comments(self, document):
        document.allow_discussion = True

        conversation = IConversation(document)
        replies = IReplies(conversation)
        comments = []
        for x in range(1, 2):
            comment = createObject("plone.Comment")
            comment.text = "Comment %d" % x
            comment = replies[replies.addComment(comment)]

            comment_replies = IReplies(comment)
            for y in range(1, 2):
                comment = createObject("plone.Comment")
                comment.text = "Comment %d.%d" % (x, y)
                comment_replies.addComment(comment)
                comments.append(comment)

        return comments

    def assert_roughly_now(self, dt):
        pydt = dt
        if isinstance(pydt, DateTime):
            pydt = pydt.asdatetime()
        elif isinstance(pydt, float):
            pydt = datetime.fromtimestamp(pydt)

        epsilon = timedelta(minutes=5)
        now = datetime.now()
        if pydt.tzinfo is not None:
            now = pydt.tzinfo.localize(now)

        upper = now + epsilon
        lower = now - epsilon

        self.assertTrue(
            lower < pydt < upper,
            "Expected %r to be now (within +/- %r). "
            "It's currently %s though, and the date time is off by %r"
            % (pydt, epsilon, now, now - pydt),
        )

    def assert_of_same_type(self, fake_datetimes, real_datetimes):
        for fake, real in zip(fake_datetimes, real_datetimes):
            # Guard against accidentally comparing the same objects
            self.assertNotEqual(fake, real)
            self.assertIsInstance(
                fake,
                real.__class__,
                "Faked static time %r is of a different "
                "type than the real %r" % (fake, real),
            )

    def test_statictime_dxcontent_created(self):
        frozen_time = datetime(1950, 7, 31, 13, 45)
        statictime = StaticTime(created=frozen_time)

        statictime.start()
        doc1 = self.create_document("doc1")
        self.assertEqual(DateTime(frozen_time), doc1.creation_date)
        fake_datetimes = [doc1.creation_date]

        statictime.stop()
        doc2 = self.create_document("doc2")
        self.assert_roughly_now(doc2.creation_date)
        real_datetimes = [doc2.creation_date]

        self.assert_of_same_type(fake_datetimes, real_datetimes)

    def test_statictime_dxcontent_modified(self):
        frozen_time = datetime(1950, 7, 31, 17, 30)
        statictime = StaticTime(modified=frozen_time)

        statictime.start()
        doc1 = self.create_document("doc1")
        self.assertEqual(DateTime(frozen_time), doc1.modification_date)
        fake_datetimes = [doc1.modification_date]

        statictime.stop()
        doc2 = self.create_document("doc2")
        self.assert_roughly_now(doc2.modification_date)
        real_datetimes = [doc2.modification_date]

        self.assert_of_same_type(fake_datetimes, real_datetimes)

    def test_statictime_comment_created(self):
        frozen_time = datetime(1950, 7, 31, 13, 45)
        statictime = StaticTime(created=frozen_time)

        statictime.start()
        doc1 = self.create_document("doc1")
        comments = self.create_comments(doc1)
        self.assertEqual(frozen_time, comments[0].creation_date)
        fake_datetimes = [comments[0].creation_date]

        statictime.stop()
        doc2 = self.create_document("doc2")
        comments = self.create_comments(doc2)
        self.assert_roughly_now(comments[0].creation_date)
        real_datetimes = [comments[0].creation_date]

        self.assert_of_same_type(fake_datetimes, real_datetimes)

    def test_statictime_comment_modified(self):
        frozen_time = datetime(1950, 7, 31, 17, 30)
        statictime = StaticTime(modified=frozen_time)

        statictime.start()
        doc1 = self.create_document("doc1")
        comments = self.create_comments(doc1)
        self.assertEqual(frozen_time, comments[0].modification_date)
        fake_datetimes = [comments[0].modification_date]

        statictime.stop()
        doc2 = self.create_document("doc2")
        comments = self.create_comments(doc2)
        self.assert_roughly_now(comments[0].modification_date)
        real_datetimes = [comments[0].modification_date]

        self.assert_of_same_type(fake_datetimes, real_datetimes)

    def test_statictime_get_info_for(self):
        frozen_time = datetime(1950, 7, 31, 17, 30)
        wftool = api.portal.get_tool("portal_workflow")
        statictime = StaticTime(modified=frozen_time)

        statictime.start()
        doc1 = self.create_document("doc1")
        api.content.transition(doc1, "publish")

        history = wftool.getInfoFor(doc1, "review_history")

        fake_datetimes = list(map(itemgetter("time"), history))
        self.assertEqual(
            fake_datetimes,
            [DateTime("1950/07/31 17:30:00 UTC"), DateTime("1950/07/31 18:30:00 UTC")],
        )

        statictime.stop()
        doc2 = self.create_document("doc2")
        api.content.transition(doc2, "publish")

        history = wftool.getInfoFor(doc2, "review_history")
        real_datetimes = list(map(itemgetter("time"), history))
        for ts in real_datetimes:
            self.assert_roughly_now(ts)

        self.assert_of_same_type(fake_datetimes, real_datetimes)

    def test_statictime_full_history(self):
        frozen_time = datetime(1950, 7, 31, 17, 30)
        statictime = StaticTime(modified=frozen_time)

        statictime.start()
        doc1 = self.create_document("doc1")
        doc1.setTitle("Current version")
        api.content.transition(doc1, "publish")
        viewlet = ContentHistoryViewlet(doc1, doc1.REQUEST, None)
        viewlet.update()

        history = viewlet.fullHistory()

        real_datetimes = list(map(itemgetter("time"), history))
        self.assertEqual(
            real_datetimes,
            [
                DateTime("1950/07/31 17:30:00 UTC"),
                -612855000.0,
                DateTime("1950/07/31 19:30:00 UTC"),
            ],
        )

        statictime.stop()
        doc2 = self.create_document("doc2")
        doc2.setTitle("Current version")
        api.content.transition(doc2, "publish")
        viewlet = ContentHistoryViewlet(doc2, doc2.REQUEST, None)
        viewlet.update()

        history = viewlet.fullHistory()

        fake_datetimes = list(map(itemgetter("time"), history))
        for ts in fake_datetimes:
            self.assert_roughly_now(ts)

        self.assert_of_same_type(fake_datetimes, real_datetimes)

    def test_statictime_lockinfo(self):
        frozen_time = datetime(1950, 7, 31, 17, 30)
        statictime = StaticTime(modified=frozen_time)

        doc1 = self.create_document("doc1")
        alsoProvides(doc1, ITTWLockable)
        lockable = ILockable(doc1)
        lockable.lock()

        statictime.start()
        lock_infos = lockable.lock_info()
        self.assertEqual(1, len(lock_infos))
        self.assertEqual(-612858600.0, lock_infos[0]["time"])
        fake_datetimes = [lock_infos[0]["time"]]

        statictime.stop()
        lock_infos = lockable.lock_info()
        self.assertEqual(1, len(lock_infos))
        self.assert_roughly_now(lock_infos[0]["time"])
        real_datetimes = [lock_infos[0]["time"]]

        self.assert_of_same_type(fake_datetimes, real_datetimes)


class TestStaticTimeWorkingCopy(unittest.TestCase):

    layer = PLONE_RESTAPI_ITERATE_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.request = self.layer["request"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()

        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        registry = getUtility(IRegistry)
        settings = registry.forInterface(IDiscussionSettings, check=False)
        settings.globally_enabled = True

        transaction.commit()

    def create_document(self, id_):
        self.portal.invokeFactory("Document", id=id_)
        document = self.portal[id_]
        document.title = "My title"
        return document

    def test_statictime_wc_created(self):
        frozen_time = datetime(1950, 7, 31, 13, 45)
        statictime = StaticTime(created=frozen_time)

        statictime.start()
        doc1 = self.create_document("doc1")

        policy = ICheckinCheckoutPolicy(doc1)
        policy.checkout(self.portal)
        baseline, working_copy = WorkingCopyInfo(doc1).get_working_copy_info()

        self.assertEqual(json_compatible(frozen_time), working_copy["created"])

        statictime.stop()



from plone import api
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from Products.CMFCore.permissions import SetOwnPassword
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import ISecuritySchema
from Products.MailHost.interfaces import IMailHost
from zope.component import getAdapter
from zope.component import getUtility

import transaction
import unittest


class TestUsersEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.mailhost = getUtility(IMailHost)

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        self.anon_api_session = RelativeSession(self.portal_url, test=self)
        self.anon_api_session.headers.update({"Accept": "application/json"})

        properties = {
            "email": "noam.chomsky@example.com",
            "username": "noamchomsky",
            "fullname": "Noam Avram Chomsky",
            "home_page": "web.mit.edu/chomsky",
            "description": "Professor of Linguistics",
            "location": "Cambridge, MA",
        }
        api.user.create(
            email="noam.chomsky@example.com",
            username="noam",
            properties=properties,
            password="password",
        )
        properties = {
            "email": "otheruser@example.com",
            "username": "otheruser",
            "fullname": "Other user",
        }
        api.user.create(
            email="otheruser@example.com",
            username="otheruser",
            properties=properties,
            password="otherpassword",
        )
        transaction.commit()

    def tearDown(self):
        self.api_session.close()
        self.anon_api_session.close()

    def test_list_users(self):
        response = self.api_session.get("/@users")

        self.assertEqual(200, response.status_code)
        self.assertEqual(4, len(response.json()))
        user_ids = [user["id"] for user in response.json()]
        self.assertIn("admin", user_ids)
        self.assertIn("test_user_1_", user_ids)
        self.assertIn("noam", user_ids)
        noam = [x for x in response.json() if x.get("username") == "noam"][0]
        self.assertEqual("noam", noam.get("id"))
        self.assertEqual(self.portal.absolute_url() + "/@users/noam", noam.get("@id"))
        self.assertEqual("noam.chomsky@example.com", noam.get("email"))
        self.assertEqual("Noam Avram Chomsky", noam.get("fullname"))
        self.assertEqual("web.mit.edu/chomsky", noam.get("home_page"))  # noqa
        self.assertEqual("Professor of Linguistics", noam.get("description"))  # noqa
        self.assertEqual("Cambridge, MA", noam.get("location"))

    def test_list_users_without_being_manager(self):
        noam_api_session = RelativeSession(self.portal_url, test=self)
        noam_api_session.headers.update({"Accept": "application/json"})
        noam_api_session.auth = ("noam", "password")

        response = noam_api_session.get("/@users")
        self.assertEqual(response.status_code, 401)
        noam_api_session.close()

    def test_list_users_as_anonymous(self):

        response = self.anon_api_session.get("/@users")
        self.assertEqual(response.status_code, 401)

    def test_add_user(self):
        response = self.api_session.post(
            "/@users",
            json={
                "username": "howard",
                "email": "howard.zinn@example.com",
                "password": "peopleshistory",
                "roles": ["Contributor"],
            },
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        howard = api.user.get(userid="howard")
        self.assertEqual("howard.zinn@example.com", howard.getProperty("email"))
        self.assertIn("Contributor", api.user.get_roles(username="howard"))

    def test_add_user_username_is_required(self):
        response = self.api_session.post("/@users", json={"password": "noamchomsky"})
        transaction.commit()

        self.assertEqual(400, response.status_code)
        self.assertTrue("Property 'username' is required" in response.text)

    def test_add_user_password_is_required(self):
        response = self.api_session.post("/@users", json={"username": "noamchomsky"})
        transaction.commit()

        self.assertEqual(400, response.status_code)
        self.assertTrue(
            ("You have to either send a " "password or sendPasswordReset")
            in response.text
        )

    def test_add_user_email_is_required_if_email_login_is_enabled(self):
        # enable use_email_as_login
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.use_email_as_login = True
        transaction.commit()
        response = self.api_session.post(
            "/@users", json={"username": "noam", "password": "secret"}
        )

        self.assertEqual(400, response.status_code)
        self.assertTrue("Property 'username' is not allowed" in response.text)

    def test_add_user_email_with_email_login_enabled(self):
        # enable use_email_as_login
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.use_email_as_login = True
        transaction.commit()
        response = self.api_session.post(
            "/@users", json={"email": "howard.zinn@example.com", "password": "secret"}
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        self.assertTrue(api.user.get(userid="howard.zinn@example.com"))

    def test_username_is_not_allowed_with_email_login_enabled(self):
        # enable use_email_as_login
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.use_email_as_login = True
        transaction.commit()
        response = self.api_session.post(
            "/@users",
            json={
                "username": "howard",
                "email": "howard.zinn@example.com",
                "password": "secret",
            },
        )
        transaction.commit()

        self.assertEqual(400, response.status_code)
        self.assertTrue("Property 'username' is not allowed" in response.text)

    def test_add_user_with_email_login_enabled(self):
        # enable use_email_as_login
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.use_email_as_login = True
        transaction.commit()
        response = self.api_session.post(
            "/@users", json={"email": "howard.zinn@example.com", "password": "secret"}
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        user = api.user.get(userid="howard.zinn@example.com")
        self.assertTrue(user)
        self.assertEqual("howard.zinn@example.com", user.getUserName())
        self.assertEqual("howard.zinn@example.com", user.getProperty("email"))

    def test_add_user_with_sendPasswordRest_sends_email(self):
        response = self.api_session.post(
            "/@users",
            json={
                "username": "howard",
                "email": "howard.zinn@example.com",
                "sendPasswordReset": True,
            },
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        msg = self.mailhost.messages[0]
        if isinstance(msg, bytes) and bytes is not str:
            # Python 3 with Products.MailHost 4.10+
            msg = msg.decode("utf-8")
        self.assertTrue("To: howard.zinn@example.com" in msg)

    def test_add_user_send_properties(self):
        response = self.api_session.post(
            "/@users",
            json={
                "username": "howard",
                "password": "secret",
                "email": "howard.zinn@example.com",
                "fullname": "Howard Zinn",
            },
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        member = api.user.get(username="howard")
        self.assertEqual(member.getProperty("fullname"), "Howard Zinn")

    def test_add_anon_user_sends_properties_are_saved(self):
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.enable_self_reg = True
        transaction.commit()

        response = self.anon_api_session.post(
            "/@users",
            json={
                "username": "howard",
                "email": "howard.zinn@example.com",
                "fullname": "Howard Zinn",
            },
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        member = api.user.get(username="howard")
        self.assertEqual(member.getProperty("fullname"), "Howard Zinn")

    def test_add_anon_no_roles(self):
        """Make sure anonymous users cannot set their own roles.
        Allowing so would make them Manager.
        """
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.enable_self_reg = True
        transaction.commit()

        response = self.anon_api_session.post(
            "/@users",
            json={
                "username": "howard",
                "email": "howard.zinn@example.com",
                "roles": ["Manager"],
            },
        )
        transaction.commit()

        self.assertEqual(400, response.status_code)
        errors = response.json()["error"]["errors"]
        fields = [x["field"] for x in errors]
        self.assertEqual(["roles"], fields)

    def test_add_user_with_uuid_as_userid_enabled(self):
        # enable use_email_as_login
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.use_email_as_login = True
        security_settings.use_uuid_as_userid = True
        transaction.commit()
        response = self.api_session.post(
            "/@users", json={"email": "howard.zinn@example.com", "password": "secret"}
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        user_id = response.json()["id"]
        user = api.user.get(userid=user_id)
        self.assertTrue(user)
        self.assertEqual("howard.zinn@example.com", user.getUserName())
        self.assertEqual("howard.zinn@example.com", user.getProperty("email"))

    def test_get_user(self):
        response = self.api_session.get("/@users/noam")

        self.assertEqual(response.status_code, 200)
        self.assertEqual("noam", response.json().get("id"))
        self.assertEqual(
            self.portal.absolute_url() + "/@users/noam", response.json().get("@id")
        )
        self.assertEqual("noam.chomsky@example.com", response.json().get("email"))
        self.assertEqual("Noam Avram Chomsky", response.json().get("fullname"))
        self.assertEqual(
            "web.mit.edu/chomsky", response.json().get("home_page")
        )  # noqa
        self.assertEqual(
            "Professor of Linguistics", response.json().get("description")
        )  # noqa
        self.assertEqual("Cambridge, MA", response.json().get("location"))

    def test_get_user_as_anonymous(self):
        response = self.anon_api_session.get("/@users/noam")
        self.assertEqual(response.status_code, 401)

    def test_get_other_user_info_when_logged_in(self):
        noam_api_session = RelativeSession(self.portal_url, test=self)
        noam_api_session.headers.update({"Accept": "application/json"})
        noam_api_session.auth = ("noam", "password")

        response = noam_api_session.get("/@users/otheruser")
        self.assertEqual(response.status_code, 401)
        noam_api_session.close()

    def test_get_search_user_with_filter(self):
        response = self.api_session.post(
            "/@users",
            json={
                "username": "howard",
                "email": "howard.zinn@example.com",
                "password": "peopleshistory",
            },
        )
        transaction.commit()
        response = self.api_session.get("/@users", params={"query": "noa"})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)
        self.assertEqual("noam", response.json()[0].get("id"))
        self.assertEqual(
            self.portal.absolute_url() + "/@users/noam", response.json()[0].get("@id")
        )
        self.assertEqual("noam.chomsky@example.com", response.json()[0].get("email"))
        self.assertEqual(
            "Noam Avram Chomsky", response.json()[0].get("fullname")
        )  # noqa

        response = self.api_session.get("/@users", params={"query": "howa"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)
        self.assertEqual("howard", response.json()[0].get("id"))

    def test_get_search_user_with_filter_as_anonymous(self):
        response = self.api_session.post(
            "/@users",
            json={
                "username": "howard",
                "email": "howard.zinn@example.com",
                "password": "peopleshistory",
            },
        )
        transaction.commit()
        response = self.anon_api_session.get("/@users", params={"query": "howa"})
        self.assertEqual(response.status_code, 401)

    def test_get_search_user_with_filter_as_unauthorized_user(self):
        response = self.api_session.post(
            "/@users",
            json={
                "username": "howard",
                "email": "howard.zinn@example.com",
                "password": "peopleshistory",
            },
        )
        transaction.commit()
        noam_api_session = RelativeSession(self.portal_url, test=self)
        noam_api_session.headers.update({"Accept": "application/json"})
        noam_api_session.auth = ("noam", "password")

        response = noam_api_session.get("/@users", params={"query": "howa"})
        self.assertEqual(response.status_code, 401)
        noam_api_session.close()

    def test_get_non_existing_user(self):
        response = self.api_session.get("/@users/non-existing-user")

        self.assertEqual(response.status_code, 404)

    def test_update_user(self):
        payload = {
            "fullname": "Noam A. Chomsky",
            "username": "avram",
            "email": "avram.chomsky@example.com",
        }
        response = self.api_session.patch("/@users/noam", json=payload)
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        noam = api.user.get(userid="noam")
        self.assertEqual("noam", noam.getUserId())  # user id never changes
        self.assertEqual("avram", noam.getUserName())
        self.assertEqual("Noam A. Chomsky", noam.getProperty("fullname"))
        self.assertEqual("avram.chomsky@example.com", noam.getProperty("email"))

    def test_user_can_update_himself(self):
        payload = {
            "fullname": "Noam A. Chomsky",
            "username": "noam",
            "email": "avram.chomsky@plone.org",
        }
        self.api_session.auth = ("noam", "password")
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 204)
        transaction.commit()

        noam = api.user.get(userid="noam")
        self.assertEqual("noam", noam.getUserId())  # user id never changes
        self.assertEqual("Noam A. Chomsky", noam.getProperty("fullname"))
        self.assertEqual("avram.chomsky@plone.org", noam.getProperty("email"))

    def test_user_can_update_himself_remove_values(self):
        payload = {
            "fullname": "Noam A. Chomsky",
            "username": "noam",
            "email": "avram.chomsky@plone.org",
            "home_page": None,
        }
        self.api_session.auth = ("noam", "password")
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 204)
        transaction.commit()

        noam = api.user.get(userid="noam")

        self.assertEqual(None, noam.getProperty("home_page"))

    def test_update_roles(self):
        self.assertNotIn("Contributor", api.user.get_roles(username="noam"))

        self.api_session.patch("/@users/noam", json={"roles": {"Contributor": True}})
        transaction.commit()
        self.assertIn("Contributor", api.user.get_roles(username="noam"))

        self.api_session.patch("/@users/noam", json={"roles": {"Contributor": False}})
        transaction.commit()
        self.assertNotIn("Contributor", api.user.get_roles(username="noam"))

    def test_update_user_password(self):
        old_password_hashes = dict(self.portal.acl_users.source_users._user_passwords)
        payload = {"password": "secret"}
        response = self.api_session.patch("/@users/noam", json=payload)
        transaction.commit()

        self.assertEqual(response.status_code, 204)

        new_password_hashes = dict(self.portal.acl_users.source_users._user_passwords)
        self.assertNotEqual(old_password_hashes["noam"], new_password_hashes["noam"])

    def test_update_portrait(self):
        payload = {
            "portrait": {
                "filename": "image.gif",
                "encoding": "base64",
                "data": "R0lGODlhAQABAIAAAP///wAAACwAAAAAAQABAAACAkQBADs=",
                "content-type": "image/gif",
            }
        }
        self.api_session.auth = ("noam", "password")
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 204)
        transaction.commit()

        user = self.api_session.get("/@users/noam").json()
        self.assertTrue(
            user.get("portrait").endswith("plone/portal_memberdata/portraits/noam")
        )

    def test_update_portrait_with_default_plone_scaling(self):
        payload = {
            "portrait": {
                "filename": "image.gif",
                "encoding": "base64",
                "data": "R0lGODlhAQABAIAAAP///wAAACwAAAAAAQABAAACAkQBADs=",
                "content-type": "image/gif",
                "scale": True,
            }
        }
        self.api_session.auth = ("noam", "password")
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 204)
        transaction.commit()

        user = self.api_session.get("/@users/noam").json()
        self.assertTrue(
            user.get("portrait").endswith("plone/portal_memberdata/portraits/noam")
        )

    def test_update_portrait_by_manager(self):
        payload = {
            "portrait": {
                "filename": "image.gif",
                "encoding": "base64",
                "data": "R0lGODlhAQABAIAAAP///wAAACwAAAAAAQABAAACAkQBADs=",
                "content-type": "image/gif",
            }
        }
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 204)
        transaction.commit()

        user = self.api_session.get("/@users/noam").json()
        self.assertTrue(
            user.get("portrait").endswith("plone/portal_memberdata/portraits/noam")
        )

    def test_delete_portrait(self):
        payload = {
            "portrait": None,
        }
        self.api_session.auth = ("noam", "password")
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 204)
        transaction.commit()

        user = self.api_session.get("/@users/noam").json()

        self.assertTrue(user.get("portrait") is None)

    def test_delete_portrait_by_manager(self):
        payload = {
            "portrait": None,
        }
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 204)
        transaction.commit()

        user = self.api_session.get("/@users/noam").json()

        self.assertTrue(user.get("portrait") is None)

    def test_update_user_with_portrait_set_without_updating_portrait(self):
        payload = {
            "portrait": {
                "filename": "image.gif",
                "encoding": "base64",
                "data": "R0lGODlhAQABAIAAAP///wAAACwAAAAAAQABAAACAkQBADs=",
                "content-type": "image/gif",
            }
        }
        self.api_session.auth = ("noam", "password")
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 204)
        transaction.commit()

        payload = {
            "fullname": "Noam A. Chomsky",
            "username": "noam",
            "email": "avram.chomsky@plone.org",
            "portrait": "http://localhost:55001/plone/portal_memberdata/portraits/noam",
        }
        self.api_session.auth = ("noam", "password")
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 204)
        transaction.commit()

        user = self.api_session.get("/@users/noam").json()
        self.assertTrue(
            user.get("portrait").endswith("plone/portal_memberdata/portraits/noam")
        )

    def test_anonymous_user_can_not_update_existing_user(self):
        payload = {
            "fullname": "Noam A. Chomsky",
            "username": "noam",
            "email": "avram.chomsky@plone.org",
        }
        self.api_session.auth = ("noam", "password")
        response = self.anon_api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 401)

    def test_user_can_not_update_another_user(self):
        payload = {
            "fullname": "Noam A. Chomsky",
            "username": "noam",
            "email": "avram.chomsky@plone.org",
        }
        self.api_session.auth = ("otheruser", "otherpassword")
        response = self.api_session.patch("/@users/noam", json=payload)

        self.assertEqual(response.status_code, 403)

    def test_user_requests_password_sends_password_via_mail(self):
        self.api_session.auth = ("noam", "password")
        payload = {}
        response = self.api_session.post("/@users/noam/reset-password", json=payload)
        transaction.commit()

        self.assertEqual(response.status_code, 200)
        # FIXME: Test that mail is sent

    def test_user_can_set_her_own_password(self):
        self.api_session.auth = ("noam", "password")
        self.portal.manage_permission(
            SetOwnPassword, roles=["Authenticated", "Manager"], acquire=False
        )
        transaction.commit()

        payload = {"old_password": "password", "new_password": "new_password"}
        response = self.api_session.post("/@users/noam/reset-password", json=payload)
        transaction.commit()

        self.assertEqual(response.status_code, 200)
        authed = self.portal.acl_users.authenticate("noam", "new_password", {})
        self.assertTrue(authed)

    def test_normal_authenticated_user_cannot_set_other_users_password(self):
        self.api_session.auth = ("noam", "password")
        self.portal.manage_permission(
            SetOwnPassword, roles=["Authenticated", "Manager"], acquire=False
        )
        transaction.commit()

        payload = {"old_password": "password", "new_password": "new_password"}
        response = self.api_session.post(
            "/@users/otheruser/reset-password", json=payload
        )
        transaction.commit()

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json()["error"]["type"], "Wrong user")

    def test_user_set_own_password_requires_set_own_password_permission(self):
        self.api_session.auth = ("noam", "password")
        self.portal.manage_permission(SetOwnPassword, roles=["Manager"], acquire=False)
        transaction.commit()

        payload = {"old_password": "password", "new_password": "new_password"}
        response = self.api_session.post("/@users/noam/reset-password", json=payload)
        transaction.commit()

        self.assertEqual(response.status_code, 403)

    def test_user_set_own_password_requires_old_and_new_password(self):
        self.api_session.auth = ("noam", "password")
        payload = {"old_password": "password"}
        response = self.api_session.post("/@users/noam/reset-password", json=payload)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()["error"]["type"], "Invalid parameters")
        payload = {"new_password": "new_password"}
        response = self.api_session.post("/@users/noam/reset-password", json=payload)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()["error"]["type"], "Invalid parameters")

    def test_user_set_own_password_checks_old_password(self):
        self.api_session.auth = ("noam", "password")
        payload = {"new_password": "new_password", "old_password": "wrong_password"}
        response = self.api_session.post("/@users/noam/reset-password", json=payload)

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json()["error"]["type"], "Wrong password")

    def test_user_set_reset_token_requires_new_password(self):
        self.api_session.auth = ("noam", "password")
        payload = {"reset_token": "abc"}
        response = self.api_session.post("/@users/noam/reset-password", json=payload)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()["error"]["type"], "Invalid parameters")

    def test_reset_with_token(self):
        reset_tool = getToolByName(self.portal, "portal_password_reset")
        reset_info = reset_tool.requestReset("noam")
        token = reset_info["randomstring"]
        transaction.commit()

        payload = {"reset_token": token, "new_password": "new_password"}
        response = self.api_session.post("/@users/noam/reset-password", json=payload)
        transaction.commit()

        self.assertEqual(response.status_code, 200)
        authed = self.portal.acl_users.authenticate("noam", "new_password", {})
        self.assertTrue(authed)

    def test_reset_with_uuid_as_userid_and_login_email_using_id(self):
        # enable use_email_as_login
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.use_email_as_login = True
        security_settings.use_uuid_as_userid = True
        transaction.commit()

        response = self.api_session.post(
            "/@users", json={"email": "howard.zinn@example.com", "password": "secret"}
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        user_id = response.json()["id"]
        user = api.user.get(userid=user_id)
        self.assertTrue(user)

        reset_tool = getToolByName(self.portal, "portal_password_reset")
        reset_info = reset_tool.requestReset(user.id)
        token = reset_info["randomstring"]
        transaction.commit()

        payload = {"reset_token": token, "new_password": "new_password"}
        response = self.api_session.post(
            f"/@users/{user.id}/reset-password", json=payload
        )

        self.assertEqual(response.status_code, 200)

    def test_reset_with_uuid_as_userid_and_login_email_using_mail(self):
        # enable use_email_as_login
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.use_email_as_login = True
        security_settings.use_uuid_as_userid = True
        transaction.commit()

        response = self.api_session.post(
            "/@users", json={"email": "howard.zinn@example.com", "password": "secret"}
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        user_id = response.json()["id"]
        user = api.user.get(userid=user_id)
        self.assertTrue(user)

        reset_tool = getToolByName(self.portal, "portal_password_reset")
        reset_info = reset_tool.requestReset(user.id)
        token = reset_info["randomstring"]
        transaction.commit()

        payload = {"reset_token": token, "new_password": "new_password"}
        response = self.api_session.post(
            f"/@users/{user.getUserName()}/reset-password", json=payload
        )

        self.assertEqual(response.status_code, 200)

    def test_reset_and_login_email_using_mail(self):
        # enable use_email_as_login
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.use_email_as_login = True
        transaction.commit()

        response = self.api_session.post(
            "/@users", json={"email": "howard.zinn@example.com", "password": "secret"}
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        user_id = response.json()["id"]
        user = api.user.get(userid=user_id)
        self.assertTrue(user)

        reset_tool = getToolByName(self.portal, "portal_password_reset")
        reset_info = reset_tool.requestReset(user.id)
        token = reset_info["randomstring"]
        transaction.commit()

        payload = {"reset_token": token, "new_password": "new_password"}
        response = self.api_session.post(
            f"/@users/{user.getUserName()}/reset-password", json=payload
        )

        self.assertEqual(response.status_code, 200)

    def test_delete_user(self):
        response = self.api_session.delete("/@users/noam")
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        self.assertEqual(None, api.user.get(userid="noam"))

    def test_delete_non_existing_user(self):
        response = self.api_session.delete("/@users/non-existing-user")
        transaction.commit()

        self.assertEqual(response.status_code, 404)

    def test_anonymous_requires_enable_self_reg(self):
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.enable_self_reg = False
        transaction.commit()

        response = self.anon_api_session.post(
            "/@users", json={"password": "noamchomsky"}
        )
        transaction.commit()

        self.assertEqual(403, response.status_code)

        security_settings.enable_self_reg = True
        transaction.commit()

        response = self.anon_api_session.post(
            "/@users",
            json={"username": "new_user", "email": "avram.chomsky@example.com"},
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)

    def test_anonymous_without_enable_user_pwd_choice_sends_mail(self):
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.enable_self_reg = True
        transaction.commit()

        response = self.anon_api_session.post(
            "/@users",
            json={"username": "new_user", "email": "avram.chomsky@example.com"},
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)
        msg = self.mailhost.messages[0]
        if isinstance(msg, bytes) and bytes is not str:
            # Python 3 with Products.MailHost 4.10+
            msg = msg.decode("utf-8")
        self.assertTrue("To: avram.chomsky@example.com" in msg)

    def test_anonymous_can_set_password_with_enable_user_pwd_choice(self):
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.enable_self_reg = True
        transaction.commit()

        response = self.anon_api_session.post(
            "/@users",
            json={
                "username": "new_user",
                "email": "avram.chomsky@example.com",
                "password": "secret",
            },
        )
        transaction.commit()

        self.assertEqual(400, response.status_code)
        self.assertTrue("Property 'password' is not allowed" in response.text)

        security_settings.enable_user_pwd_choice = True
        transaction.commit()

        response = self.anon_api_session.post(
            "/@users",
            json={
                "username": "new_user",
                "email": "avram.chomsky@example.com",
                "password": "secret",
            },
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)

    def test_anonymous_with_enable_user_pwd_choice_doent_send_email(self):
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.enable_self_reg = True
        security_settings.enable_user_pwd_choice = True
        transaction.commit()

        response = self.anon_api_session.post(
            "/@users",
            json={
                "username": "new_user",
                "email": "avram.chomsky@example.com",
                "password": "secret",
            },
        )
        transaction.commit()

        self.assertEqual(self.mailhost.messages, [])
        self.assertEqual(201, response.status_code)

    def test_anonymous_with_enable_user_sets_only_member_role(self):
        security_settings = getAdapter(self.portal, ISecuritySchema)
        security_settings.enable_self_reg = True
        security_settings.enable_user_pwd_choice = True
        transaction.commit()

        response = self.anon_api_session.post(
            "/@users",
            json={
                "username": "new_user",
                "email": "avram.chomsky@example.com",
                "password": "secret",
            },
        )

        response = response.json()
        self.assertIn("Member", response["roles"])
        self.assertEqual(1, len(response["roles"]))

    def test_add_user_no_roles_sets_member_as_sensible_default(self):
        response = self.api_session.post(
            "/@users",
            json={
                "username": "howard",
                "email": "howard.zinn@example.com",
                "password": "peopleshistory",
            },
        )
        transaction.commit()

        self.assertEqual(201, response.status_code)

        response = response.json()

        self.assertIn("Member", response["roles"])
        self.assertEqual(1, len(response["roles"]))



from Acquisition import aq_parent
from plone.app.linkintegrity.exceptions import LinkIntegrityNotificationException
from plone.restapi.services import Service


class ContentDelete(Service):
    """Deletes a content object."""

    def reply(self):

        parent = aq_parent(self.context)
        try:
            parent.manage_delObjects([self.context.getId()])
        except LinkIntegrityNotificationException:
            pass

        return self.reply_no_content()
